"""
Class members of AbsOpBase like in_dtypes, out_dtypes, etc. are just the superset
of the valid domain that do not take the upcoming inference engine's operator
availability into account. For example, older versions of TVM may not support trilu
and TensorRT does not accept a 2D Pool whose kernel is larger than 300. Therefore,
to narrow those specifications, we look at the following methods:
- Identifier: Model and BackendFactory
1. Single-operator specification testing: Iterate over possible operator instances
   of available `data types` (and `ranks`), and try to kick out failing ones (loggings
   will be kept to see if it is just an unimplemented feature or a bug).
2. Constraint extension: for a BackendFactory, it can add more constraints to an operator.
   This is useful for cases like TensorRT where the kernel size is limited to < 300.
- HOWTO:
1. Single-operator potential data types (and ranks) are serializable as a JSON file.
2. Constraint extensions are defined as Python codes. That said, we can just overload the
   operator specifactions in backend (see nnsmith/abstract/extension).
"""


import os
import tempfile
from copy import deepcopy
from dataclasses import dataclass
from os import PathLike
from typing import Dict, List, Optional, Type

import z3
from appdirs import user_cache_dir
from omegaconf import OmegaConf

from nnsmith import __version__
from nnsmith.abstract.dtype import DType
from nnsmith.abstract.op import (
    AbsOpBase,
    AbsTensor,
    Constant,
    Input,
    Placeholder,
    concretize_op,
)
from nnsmith.backends import BackendFactory
from nnsmith.gir import GraphIR, InstExpr
from nnsmith.logging import DTEST_LOG
from nnsmith.materialize import Model, TestCase

NNSMITH_CACHE_DIR = user_cache_dir(f"nnsmith-{__version__}")


def get_cache_name(model_cls: Type[Model], factory: BackendFactory) -> str:
    if factory is None:
        return f"{model_cls.__name__}_exportable"
    return (
        f"{model_cls.__name__}_{factory.system_name}_{factory.version}_{factory.target}"
    )


@dataclass
class OpConfig:
    in_dtypes: List[List[DType]]
    out_dtypes: List[List[DType]]


def _make_single_op_irs(
    op: AbsOpBase, ishapes, available_idtypes
):  # List<Tup<DTypeComb,DTypeComb,GraphIR>>
    """Given a concretized compute op, return an GraphIR for it."""
    ir_list = []
    for idtype_group in available_idtypes:
        ir = GraphIR()
        inputs = []

        for ishape, idtype in zip(ishapes, idtype_group):
            input_op = Placeholder(AbsTensor(ishape, idtype)).to_input()
            inst = ir.add_inst(InstExpr(op=input_op, args=[]))
            inputs.append(inst.retval())

        this_op = deepcopy(op)
        itensors = [ir.vars[vname] for vname in inputs]
        otensors = this_op.checked_type_transfer(itensors)
        this_op.bind_input_like(itensors)
        this_op.bind_output_like(otensors)

        ir.add_inst(InstExpr(op=this_op, args=inputs))

        ir_list.append((idtype_group, tuple([out.dtype for out in otensors]), ir))
    return ir_list


def infer_topset_from_scratch(
    model_cls: Model, factory: Optional[BackendFactory], op_types=None
) -> Dict[str, OpConfig]:
    if op_types is None:
        op_types = model_cls.operators()

    topset = {}

    n_ops = len(op_types)
    for idx, node_t in enumerate(op_types):
        if node_t is Input or node_t is Constant:
            continue

        DTEST_LOG.info(f"[{idx + 1} / {n_ops}] ===> Trying {node_t}")

        available_idtypes = node_t.in_dtypes

        if not available_idtypes:
            continue

        op_param_n = node_t.get_num_var_param()
        op_params = [z3.Int("v%s-%s" % (idx, k)) for k in range(op_param_n)]
        op = node_t(*op_params)

        solver = z3.Solver()

        inputs = []
        for i, ranks in enumerate(op.inp_ranks):
            if op.same_inp_dims and inputs:
                rank = inputs[0].ndims
            else:  # FIXME(@ganler): consider rank check over scalar & non-scalar.
                rank = min(ranks)
            shape = AbsTensor(
                shape=[z3.Int("s%s" % (k)) for k in range(rank)],
                dtype=available_idtypes[0][i],
            )
            inputs.append(shape)
            solver.add(*shape.gt_zero())
            solver.add(*[s < 64 for s in shape.shape])

        solver.add(*op.checked_requires(inputs))

        # solve
        assert solver.check() == z3.sat, f"Cannot solve the problem in {node_t}"
        m = solver.model()

        # rewrite
        concrete_op = concretize_op(op, m)
        concre_input_shapes = []
        for inp in inputs:
            shape = []
            for s in inp.shape:
                shape.append(m.eval(s).as_long())
            concre_input_shapes.append(shape)

        single_op_irs = _make_single_op_irs(
            concrete_op, concre_input_shapes, available_idtypes
        )

        # filter out unsupported dtypes by model.
        single_op_irs = [
            sset
            for sset in single_op_irs
            if set(model_cls.skip_dtypes()).isdisjoint(sset[0] + sset[1])
        ]

        if factory:
            single_op_irs = [
                sset
                for sset in single_op_irs
                if set(factory.skip_dtypes()).isdisjoint(sset[0] + sset[1])
            ]

        op_itypes = set()
        op_otypes = set()

        for itypes, otypes, sched in single_op_irs:
            model = model_cls.from_gir(sched)

            if factory:
                # Test compilation + simple inference;
                out = factory.make_testcase(model)
                if isinstance(out, TestCase):  # Pass
                    DTEST_LOG.info(
                        f"=====> [Success] at {concrete_op}({itypes}) => {otypes}"
                    )
                    op_itypes.add(itypes)
                    op_otypes.add(otypes)
                else:  # Fail
                    DTEST_LOG.warning(
                        f"=====> [Failure] at {concrete_op}({itypes}) => {otypes}"
                    )
                    DTEST_LOG.debug(f"{out.log}")
            else:  # Test model dumping
                with tempfile.TemporaryDirectory() as tmpdirname:
                    try:
                        model.make_oracle()  # try-run.
                        model_path = os.path.join(
                            tmpdirname, model.name_prefix() + model.name_suffix()
                        )
                        model.dump(model_path)
                        DTEST_LOG.info(
                            f"=====> [Success] at {concrete_op}({itypes}) => {otypes}"
                        )
                        op_itypes.add(itypes)
                        op_otypes.add(otypes)
                    except Exception as e:
                        DTEST_LOG.warning(
                            f"=====> [Failure] at {concrete_op}({itypes}) => {otypes}"
                        )
                        DTEST_LOG.debug(f"{e}")

        if op_itypes:
            topset[op.name()] = OpConfig(
                in_dtypes=list(op_itypes), out_dtypes=list(op_otypes)
            )

    return topset


def load_topset(topset_path) -> Dict[str, OpConfig]:
    conf = OmegaConf.load(topset_path)["topset"]
    ret = {}
    # cvt str -> DType
    for k, v in conf.items():
        ret[k] = OpConfig(
            in_dtypes=[tuple([DType[t] for t in dtypes]) for dtypes in v["in_dtypes"]],
            out_dtypes=[
                tuple([DType[t] for t in dtypes]) for dtypes in v["out_dtypes"]
            ],
        )
    return ret


def dump_topset(topset: Dict[str, OpConfig], path: PathLike):
    OmegaConf.save({"topset": topset}, path)


def auto_opconfig(
    model_cls: Model, factory: Optional[BackendFactory]
) -> Dict[str, OpConfig]:
    cache_path = os.path.join(
        NNSMITH_CACHE_DIR, get_cache_name(model_cls, factory) + ".yaml"
    )
    # mkdir -p NNSMITH_CACHE_DIR
    if not os.path.exists(NNSMITH_CACHE_DIR):
        os.makedirs(NNSMITH_CACHE_DIR)
    if os.path.exists(cache_path):
        DTEST_LOG.info(f"Loading topset from {cache_path}.")
        DTEST_LOG.info(
            "To regenerate the topset, delete the cache file above and restart."
        )
        return load_topset(cache_path)
    else:
        DTEST_LOG.info(f"Inferring topset from scratch and cache it to {cache_path}.")
        opset = infer_topset_from_scratch(model_cls, factory)
        dump_topset(opset, cache_path)
        return opset


def auto_opset(
    model_cls: Type[Model],
    factory: Optional[BackendFactory] = None,
    vulops: bool = False,
) -> List[Type[AbsOpBase]]:
    # None means only test model exportation.
    topset_config = auto_opconfig(model_cls, factory)
    opset = []
    for op in model_cls.operators():
        if op.name() not in topset_config or (vulops == False and op.limit_domain):
            continue

        op.in_dtypes = topset_config[op.name()].in_dtypes
        op.out_dtypes = topset_config[op.name()].out_dtypes
        opset.append(op)

    return opset



import hydra
from omegaconf import DictConfig

from nnsmith.backends import BackendFactory
from nnsmith.materialize import Model
from nnsmith.narrow_spec import auto_opconfig


@hydra.main(version_base=None, config_path="../config", config_name="main")
def main(cfg: DictConfig):
    backend_cfg = cfg["backend"]
    if backend_cfg["type"] is not None:
        factory = BackendFactory.init(
            name=backend_cfg["type"],
            target=backend_cfg["target"],
            optmax=backend_cfg["optmax"],
        )
    else:
        factory = None
    model_type = Model.init(cfg["model"]["type"], backend_target=backend_cfg["target"])
    auto_opconfig(model_type, factory)


if __name__ == "__main__":
    main()



from inspect import signature
from types import FunctionType
from typing import Set, Type

from nnsmith.materialize import BugReport, Stage, Symptom

FILTERS = {}


class filter:
    def __init__(self, name):
        self.name = name

    def __call__(self, fn_or_cls):
        assert self.name not in FILTERS, f"Filter {self.name} already exists."
        if isinstance(fn_or_cls, Type):  # Class checks
            assert not signature(
                fn_or_cls
            ).parameters, f"filter class {fn_or_cls.__name__} (aka {self.name}) should not have any parameters."
            caller_sig = signature(fn_or_cls.__call__).parameters
            assert (
                caller_sig["self"] and len(caller_sig) == 2
            ), f"filter class {fn_or_cls.__name__} (aka {self.name}) should implement __call__(self, report: BugReport)."
        elif isinstance(fn_or_cls, FunctionType):  # Function checks
            caller_sig = signature(fn_or_cls).parameters
            assert (
                len(caller_sig) == 1
            ), f"filter function {fn_or_cls.__name__} (aka {self.name}) should implement __call__(report: BugReport)."
        else:
            raise ValueError(
                f"filter {fn_or_cls} (aka {self.name}) should be a class or function."
            )

        FILTERS[self.name] = fn_or_cls
        return fn_or_cls


@filter("nan")
def filter_nan(report: BugReport) -> bool:  # True means filter;
    if report.symptom != Symptom.INCONSISTENCY or report.stage != Stage.VERIFICATION:
        return False

    # numpy.assert_allclose style.
    # TODO(ganler): can we use more well-formed checking? say directly checking the results?
    return (
        "nan location mismatch" in report.log
        or "-9223372036854775808" in report.log  # tf.cast(nan, int) is UB.
        or "-2147483648" in report.log
    )


@filter("inf")
def filter_inf(report: BugReport) -> bool:
    if report.symptom != Symptom.INCONSISTENCY or report.stage != Stage.VERIFICATION:
        return False

    # numpy.assert_allclose style.
    return "inf" in report.log.replace("Max relative difference: inf", "")


@filter("dup")  # duplicate
class FilterDup:
    def __init__(self):
        self.seen: Set[int] = set()

    def __call__(self, report: BugReport) -> bool:
        if (
            report.symptom != Symptom.EXCEPTION
            and report.symptom != Symptom.INCONSISTENCY
        ):
            return False  # don't filter bugs other than inconsistency/exception

        str_hash = hash(report.log)
        if str_hash in self.seen:
            return True

        self.seen.add(str_hash)
        return False


# You can patch your own filters!



import pytest

from nnsmith.abstract.arith import nnsmith_lt
from nnsmith.abstract.extension import activate_ext, patch_requires
from nnsmith.abstract.op import Constant, Input, NCHWConv2d, ReLU
from nnsmith.graph_gen import model_gen


def test_only_conv_relu():
    gen = model_gen(
        opset=[ReLU, NCHWConv2d],
        max_nodes=5,
        rank_choices=(4,),
        dtype_choices=("float32",),
    )

    ir = gen.make_concrete()

    for inst in ir.insts:
        assert type(inst.iexpr.op) in [ReLU, Input, Constant, NCHWConv2d]


def test_constrain_conv_ksize():
    @patch_requires("global", "core.NCHWConv2d")
    def limit_conv2d(self, _):
        # let the kernels to be > 3
        return [nnsmith_lt(3, self.kernel_h_size), nnsmith_lt(3, self.kernel_w_size)]

    opset = [ReLU, NCHWConv2d]
    activate_ext(opset)
    gen = model_gen(
        opset=opset,
        max_nodes=5,
        rank_choices=(4,),
        dtype_choices=("float32",),
    )

    ir = gen.make_concrete()
    for inst in ir.insts:
        assert type(inst.iexpr.op) in [ReLU, Input, Constant, NCHWConv2d]
        if isinstance(inst.iexpr.op, NCHWConv2d):
            assert inst.iexpr.op.kernel_h_size > 3
            assert inst.iexpr.op.kernel_w_size > 3



from dataclasses import dataclass
from typing import List

import numpy as np
import onnx
import pycuda.driver as cuda
import tensorrt as trt
from multipledispatch import dispatch
from pycuda.driver import DeviceAllocation

from nnsmith.abstract.arith import *
from nnsmith.abstract.dtype import DType
from nnsmith.abstract.extension import patch_requires
from nnsmith.abstract.op import AbsOpBase
from nnsmith.abstract.tensor import AbsTensor
from nnsmith.backends import BackendFactory
from nnsmith.materialize.onnx import ONNXModel


@dataclass
class HostDeviceMem:
    host: np.ndarray
    device: DeviceAllocation


class TRTFactory(BackendFactory):
    def __init__(self, target="cuda", optmax=True, **kwargs):
        super().__init__(target, optmax, **kwargs)

        if target != "cuda":
            raise ValueError("TensorRT backend only supports GPU!")

        if optmax is False:
            # TODO(@ganler): support non-optimized TensorRT by using performing
            # inference over a model that marks all nodes as outputs.
            raise ValueError("There is not O0 mode for TensorRT so far.")

    @property
    def system_name(self) -> str:
        return "tensorrt"

    @property
    def version(self) -> str:
        return trt.__version__

    @dispatch(ONNXModel)
    def make_backend(self, model: ONNXModel):
        import pycuda.autoinit

        onnx_model = model.native_model
        engine = self.build_engine_onnx(onnx_model)

        def closure(inputs):
            (
                trt_inputs,
                trt_outputs,
                trt_bindings,
                stream,
                onames,
                name2idx,
            ) = self.allocate_buffers(engine)
            context = engine.create_execution_context()

            for iname in inputs:
                np.copyto(
                    trt_inputs[name2idx[iname]].host,
                    inputs[iname]
                    .astype(trt.nptype(engine.get_binding_dtype(iname)))
                    .ravel(),
                )

            trt_concrete_outputs = self.do_inference_v2(
                context,
                bindings=trt_bindings,
                inputs=trt_inputs,
                outputs=trt_outputs,
                stream=stream,
            )

            return {
                n: v.reshape(engine.get_binding_shape(n))
                for n, v in zip(onames, trt_concrete_outputs)
            }

        return closure

    @staticmethod
    def build_engine_onnx(onnx_model):
        builder = trt.Builder(trt.Logger(trt.Logger.WARNING))
        network = builder.create_network(
            1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
        )
        config = builder.create_builder_config()
        config.set_memory_pool_limit(trt.MemoryPoolType.WORKSPACE, 2 << 30)
        parser = trt.OnnxParser(network, trt.Logger(trt.Logger.WARNING))
        # Load the Onnx model and parse it in order to populate the TensorRT network.
        if not parser.parse(onnx._serialize(onnx_model)):
            error_msg = ""
            for error in range(parser.num_errors):
                error_msg += str(parser.get_error(error))
            raise RuntimeError(error_msg)
        engine_bytes = builder.build_serialized_network(network, config)
        return trt.Runtime(trt.Logger(trt.Logger.WARNING)).deserialize_cuda_engine(
            engine_bytes
        )

    @staticmethod
    def allocate_buffers(engine):
        inputs = []
        outputs = []
        bindings = []
        stream = cuda.Stream()
        onames = []
        name2idx = {}
        for idx, binding in enumerate(engine):
            name2idx[binding] = idx
            size = trt.volume(engine.get_binding_shape(binding)) * engine.max_batch_size
            dtype = trt.nptype(engine.get_binding_dtype(binding))
            # Allocate host and device buffers
            host_mem = cuda.pagelocked_empty(size, dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)
            # Append the device buffer to device bindings.
            bindings.append(int(device_mem))
            # Append to the appropriate list.
            if engine.binding_is_input(binding):
                inputs.append(HostDeviceMem(host_mem, device_mem))
            else:
                outputs.append(HostDeviceMem(host_mem, device_mem))
                onames.append(binding)
        return inputs, outputs, bindings, stream, onames, name2idx

    @staticmethod
    def do_inference_v2(context, bindings, inputs, outputs, stream):
        # Transfer input data to the GPU.
        [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
        # Run inference.
        context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)
        # Transfer predictions back from the GPU.
        [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
        # Synchronize the stream
        stream.synchronize()
        # Return only the host outputs.
        return [out.host for out in outputs]

    @classmethod
    def skip_dtypes(cls) -> List[DType]:
        # TRT will truncate f64 -> f32 and i64 -> i32
        return [DType.float64, DType.int64]


@patch_requires(TRTFactory.system_name, "core.Pool2d")
def RulePool2d(self: AbsOpBase, _: List[AbsTensor]) -> List[Union[z3.BoolRef, bool]]:
    return [nnsmith_lt(nnsmith_mul(self.kernel_h_size, self.kernel_w_size), 10000)]



from dataclasses import dataclass
from typing import Callable, Dict, List, cast

import tensorflow as tf

from nnsmith.abstract.op import AbsOpBase, Input
from nnsmith.error import SanityCheck
from nnsmith.gir import GraphIR
from nnsmith.logging import TF_LOG
from nnsmith.materialize.tensorflow.forward import forward_fn


@dataclass
class Instr:
    fwd_fn: Callable
    inp_keys: List[str]
    out_keys: List[str]


class TFNet(tf.Module):
    """A TensorFlow network whose computation is defined by a GraphIR."""

    def __init__(self, ir: GraphIR) -> None:
        """Build a TensorFlow model from GraphIR

        Args:
            ir (GraphIR): minimal information for constructing a concrete graph.
        """
        super().__init__()
        self.ir: GraphIR = ir
        self.mlist: List[Callable] = []
        self.instructions: List[Instr] = []

        for inst in self.ir.insts:
            if not isinstance(inst.iexpr.op, Input):
                op = cast(AbsOpBase, inst.iexpr.op)
                fwd_fn = forward_fn(op)
                SanityCheck.true(fwd_fn is not None, f"Bad impl for {inst.iexpr.op}")
                if isinstance(fwd_fn, tf.Module):
                    self.mlist.append(fwd_fn)  # Add tf.Module to track its parameters
                self.instructions.append(Instr(fwd_fn, inst.iexpr.args, inst.retvals()))

    @tf.function
    def __call__(self, *args, **kwargs) -> Dict[str, tf.Tensor]:
        return self.__forward(*args, **kwargs)

    @tf.function
    def call_by_dict(self, x: Dict[str, tf.Tensor]) -> Dict[str, tf.Tensor]:
        return self.__forward(**x)

    @tf.function
    def __forward(self, *args, **kwargs) -> Dict[str, tf.Tensor]:
        mode = "Running Eagerly" if tf.executing_eagerly() else "Tracing"
        TF_LOG.debug(f"{mode} with JIT config: {tf.config.optimizer.get_jit()}")

        key2tensor: Dict[str, tf.Tensor] = {}
        if len(args) == len(self.ir.input_var()):
            for i, key in enumerate(self.ir.input_var()):
                key2tensor[key] = args[i]
        elif len(kwargs) == len(self.ir.input_var()):
            for i, key in enumerate(self.ir.input_var()):
                key2tensor[key] = kwargs[key]
        else:
            raise ValueError("Either user args only or kwargs only")

        for instr in self.instructions:
            # get inputs
            inp_tensors = [key2tensor[key] for key in instr.inp_keys]

            # forward
            out_tensors = instr.fwd_fn(
                *inp_tensors
            )  # TODO Colin when it can return a list?
            if not isinstance(out_tensors, list):
                out_tensors = [out_tensors]

            # store outputs
            for i_out, out_key in enumerate(instr.out_keys):
                key2tensor[out_key] = out_tensors[i_out]

        return {k: key2tensor[k] for k in self.ir.leaf_var()}



from typing import Dict

import tensorflow as tf  # type: ignore
from multipledispatch import dispatch

from nnsmith.backends.factory import BackendCallable, BackendFactory
from nnsmith.materialize.tensorflow import (
    EagerModeCtx,
    TFModel,
    np_dict_from_tf,
    tf_dict_from_np,
)


class XLAFactory(BackendFactory):
    def __init__(self, target="cpu", optmax: bool = False):
        super().__init__(target, optmax)

        if self.target == "cpu":
            self.device = tf.device(tf.config.list_logical_devices("CPU")[0].name)
        elif self.target == "cuda":
            self.device = tf.device(tf.config.list_logical_devices("GPU")[0].name)
        else:
            raise ValueError(
                f"Unknown device: {self.target}. Only `cpu` and `cuda` are supported."
            )

    @property
    def system_name(self) -> str:
        return "xla"

    @property
    def version(self) -> str:
        return tf.__version__

    @dispatch(TFModel)
    def make_backend(self, model: TFModel) -> BackendCallable:
        with self.device, EagerModeCtx(False):
            compiled = tf.function(jit_compile=True)(model.concrete_net())

        def closure(inputs: Dict[str, tf.Tensor]) -> Dict[str, tf.Tensor]:
            with self.device, EagerModeCtx(False):
                result = np_dict_from_tf(compiled(**tf_dict_from_np(inputs)))
            return result

        return closure



import logging
import os
import random
import time

import hydra
from omegaconf import DictConfig

from nnsmith.abstract.extension import activate_ext
from nnsmith.backends.factory import BackendFactory
from nnsmith.graph_gen import SymbolicGen, model_gen, viz
from nnsmith.logging import MGEN_LOG
from nnsmith.materialize import Model, TestCase
from nnsmith.narrow_spec import auto_opset
from nnsmith.util import hijack_patch_requires, mkdir, op_filter


@hydra.main(version_base=None, config_path="../config", config_name="main")
def main(cfg: DictConfig):
    # Generate a random ONNX model
    # TODO(@ganler): clean terminal outputs.
    mgen_cfg = cfg["mgen"]

    seed = random.getrandbits(32) if mgen_cfg["seed"] is None else mgen_cfg["seed"]

    MGEN_LOG.info(f"Using seed {seed}")

    # TODO(@ganler): skip operators outside of model gen with `cfg[exclude]`
    model_cfg = cfg["model"]
    ModelType = Model.init(model_cfg["type"], backend_target=cfg["backend"]["target"])
    ModelType.add_seed_setter()

    if cfg["backend"]["type"] is not None:
        factory = BackendFactory.init(
            cfg["backend"]["type"],
            target=cfg["backend"]["target"],
            optmax=cfg["backend"]["optmax"],
        )
    else:
        factory = None

    # GENERATION
    opset = auto_opset(ModelType, factory, vulops=mgen_cfg["vulops"])
    opset = op_filter(opset, mgen_cfg["include"], mgen_cfg["exclude"])
    hijack_patch_requires(mgen_cfg["patch_requires"])
    activate_ext(opset=opset, factory=factory)

    tgen_begin = time.time()
    gen = model_gen(
        opset=opset,
        method=mgen_cfg["method"],
        seed=seed,
        max_elem_per_tensor=mgen_cfg["max_elem_per_tensor"],
        max_nodes=mgen_cfg["max_nodes"],
        timeout_ms=mgen_cfg["timeout_ms"],
        rank_choices=mgen_cfg["rank_choices"],
        dtype_choices=mgen_cfg["dtype_choices"],
    )
    tgen = time.time() - tgen_begin

    if isinstance(gen, SymbolicGen):
        MGEN_LOG.info(
            f"{len(gen.last_solution)} symbols and {len(gen.solver.assertions())} constraints."
        )

        if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
            MGEN_LOG.debug("solution:" + ", ".join(map(str, gen.last_solution)))

    # MATERIALIZATION
    tmat_begin = time.time()
    ir = gen.make_concrete()

    MGEN_LOG.info(
        f"Generated DNN has {ir.n_var()} variables and {ir.n_compute_inst()} operators."
    )

    mkdir(mgen_cfg["save"])
    if cfg["debug"]["viz"]:
        fmt = cfg["debug"]["viz_fmt"].replace(".", "")
        viz(ir, os.path.join(mgen_cfg["save"], f"graph.{fmt}"))

    model = ModelType.from_gir(ir)
    model.refine_weights()  # either random generated or gradient-based.
    oracle = model.make_oracle()
    tmat = time.time() - tmat_begin

    tsave_begin = time.time()
    testcase = TestCase(model, oracle)
    testcase.dump(root_folder=mgen_cfg["save"])
    tsave = time.time() - tsave_begin

    MGEN_LOG.info(
        f"Time:  @Generation: {tgen:.2f}s  @Materialization: {tmat:.2f}s  @Save: {tsave:.2f}s"
    )


if __name__ == "__main__":
    main()



import json
import os
import pickle
from abc import ABC, abstractmethod
from enum import Enum
from os import PathLike
from typing import Any, Dict, List, Type, TypeVar

import numpy as np
from multipledispatch import dispatch

from nnsmith.abstract.dtype import DType
from nnsmith.abstract.op import AbsOpBase, Constant
from nnsmith.abstract.tensor import AbsTensor
from nnsmith.error import SanityCheck
from nnsmith.gir import GraphIR
from nnsmith.util import HAS_PYGRAPHVIZ, viz_dot


def framework_operator_impl(
    framework_realizable_ops: List[Type[AbsOpBase]],
    all_framework_ops: List[Type[AbsOpBase]],
    op_type: AbsOpBase,
    *args,
    **kwargs,
):
    """When implementing `forward_fn` for an operator class, add this operator into all_framework_ops list.

    Usage:
        In `forward.py`, define `operator_impl = partial(framework_operator_impl, FW_REALIZABLE_OPS, ALL_FM_OPS)`.
        Then add `@operator_impl(OpClass)` when implementing `forward_fn` for `OpClass`.

    Args:
        framework_realizable_ops (List[Type[AbsOpBase]]): all realizable ops in the framework. Usually it can be obtained by FULL_OPERATOR_SETS["core"].union(FULL_OPERATOR_SETS["framework_name"])
        all_framework_ops (List[Type[AbsOpBase]]): list of operator classes that are implemented `forward_fn` in the framework.
        op_type (AbsOpBase): operator class
    """
    SanityCheck.true(
        issubclass(op_type, AbsOpBase),
        f"Decorator operator_impl takes AbsOpBase subclass, but got {op_type}",
    )
    if op_type is not Constant:  # Constant comes from placeholder.
        dispatchables = [
            rtype for rtype in framework_realizable_ops if issubclass(rtype, op_type)
        ]
        for rtype in dispatchables:
            all_framework_ops.append(rtype)

        SanityCheck.true(
            len(dispatchables) != 0,
            f"Decorator operator_impl only take types decorated by `mark_realize`, but got {op_type}",
        )
    return dispatch(op_type, *args, **kwargs)


class Oracle:
    def __init__(
        self,
        input: Dict[str, np.ndarray],
        output: Dict[str, np.ndarray],
        provider: str = "unknown",
    ) -> None:
        self.input: Dict[str, np.ndarray] = input
        self.output: Dict[str, np.ndarray] = output
        self._provider = provider

    def __repr__(self) -> str:
        return f"input={self.input}, output={self.output}"

    @staticmethod
    def name() -> str:
        return "oracle.pkl"

    @property
    def provider(self) -> str:
        return self._provider

    def dump(self, path: PathLike) -> None:
        with open(path, "wb") as f:
            to_dump = {
                "input": self.input,
                "output": self.output,
                "provider": self.provider,
            }
            pickle.dump(to_dump, f)

    @staticmethod
    def load(path: PathLike) -> "Oracle":
        with open(path, "rb") as f:
            to_load = pickle.load(f)
            return Oracle(to_load["input"], to_load["output"], to_load["provider"])


MT = TypeVar("MT", bound="Model")


class Model(ABC):
    @property
    @abstractmethod
    def input_like(self) -> Dict[str, AbsTensor]:
        pass

    @property
    @abstractmethod
    def output_like(self) -> Dict[str, AbsTensor]:
        pass

    @classmethod
    @abstractmethod
    def from_gir(cls: Type[MT], ir: GraphIR, **kwargs) -> MT:
        pass

    @classmethod
    @abstractmethod
    def load(cls: Type[MT], path: PathLike) -> MT:
        pass

    @abstractmethod
    def dump(self, path: PathLike) -> None:
        pass

    @property
    @abstractmethod
    def native_model(self) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def name_suffix() -> str:
        """Suffix of the model file.

        Returns:
            str: Model suffix such as ".onnx".

        Note:
            Model as a folder can be "".
        """
        pass

    @abstractmethod
    def refine_weights(self) -> None:
        pass

    @abstractmethod
    def make_oracle(self) -> Oracle:
        pass

    @staticmethod
    @abstractmethod
    def operators() -> List[Type[AbsOpBase]]:
        pass

    @property
    def version(self) -> str:
        return "unknown"

    @staticmethod
    def name_prefix() -> str:
        return "model"

    @staticmethod
    def operators() -> List[Type[AbsOpBase]]:
        pass

    @staticmethod
    def add_seed_setter() -> None:
        pass

    def attach_viz(self, ir: GraphIR) -> None:
        if HAS_PYGRAPHVIZ:
            self.dotstring = ir.to_dot()

    def dump_viz(self, path: PathLike) -> None:
        viz_dot(self.dotstring, path)

    @staticmethod
    def skip_dtypes() -> List[DType]:
        return []

    @staticmethod
    def init(name, backend_target=None) -> Type["Model"]:
        if name is None:
            raise ValueError(
                "Model type cannot be None. Use `model.type=[torch|onnx|tensorflow]`."
            )

        if name == "torch":
            from nnsmith.materialize.torch import TorchModelCPU, TorchModelCUDA

            if backend_target == "gpu" or backend_target == "cuda":
                return TorchModelCUDA
            return TorchModelCPU
        elif name == "onnx":
            # device agnoistic
            from nnsmith.materialize.onnx import ONNXModelCPU, ONNXModelCUDA

            if backend_target == "gpu" or backend_target == "cuda":
                return ONNXModelCUDA
            return ONNXModelCPU
        elif name == "tensorflow":
            from nnsmith.materialize.tensorflow import TFModelCPU, TFModelCUDA

            if backend_target == "gpu" or backend_target == "cuda":
                # XLA must align device location of eager mode execution.
                return TFModelCUDA
            return TFModelCPU

        raise ValueError(
            f"Unsupported: ModelType={name} for BackendTarget={backend_target}"
        )


class TestCase:
    def __init__(self, model: Model, oracle: Oracle):
        self.oracle = oracle
        self.model = model

    @staticmethod
    def load(
        model_type: Type[Model], root_folder: str, allow_no_oracle=False
    ) -> "TestCase":
        model_path = os.path.join(
            root_folder, model_type.name_prefix() + model_type.name_suffix()
        )
        model = model_type.load(model_path)

        assert allow_no_oracle or os.path.exists(
            os.path.join(root_folder, Oracle.name())
        ), "Oracle is not found or auto-generated when allow_no_oracle is True."

        oracle = None
        if os.path.exists(os.path.join(root_folder, Oracle.name())):
            oracle_path = os.path.join(root_folder, Oracle.name())
            oracle = Oracle.load(oracle_path)

        return TestCase(model, oracle)

    def dump(self, root_folder: str):
        if self.model and hasattr(self.model, "dotstring"):
            self.model.dump_viz(os.path.join(root_folder, "graph.png"))

        self.model.dump(
            os.path.join(
                root_folder, self.model.name_prefix() + self.model.name_suffix()
            )
        )
        self.oracle.dump(os.path.join(root_folder, Oracle.name()))


class Symptom(Enum):
    EXCEPTION = "exception"
    SEGFAULT = "segfault"
    INCONSISTENCY = "inconsistency"
    TIMEOUT = "timeout"
    WORSE_PERF = "worse_perf"


class Stage(Enum):
    COMPILATION = "compilation"
    EXECUTION = "execution"
    VERIFICATION = "verification"


class BugReport(ABC):
    def __init__(
        self,
        testcase: TestCase,
        symptom: Symptom,
        stage: Stage,
        system: str,
        version: str = None,
        version_id: str = None,
        log: str = None,
    ):
        self.testcase = testcase
        self.symptom = symptom
        self.system = system
        self.stage = stage
        self.version = version
        self.version_id = version_id
        self.log = log

    @staticmethod
    def error_msg_name():
        return "err.log"

    def __repr__(self) -> str:
        return f"{self.system} {self.symptom.value} in {self.stage.value}\n{self.log}"

    def dump(self, root_folder: str):
        # create folder if not exists
        if not os.path.exists(root_folder):
            os.makedirs(root_folder)

        # model*
        # oracle.pkl
        self.testcase.dump(root_folder)
        # err.log
        with open(os.path.join(root_folder, self.error_msg_name()), "w") as f:
            f.write(self.log)
        # meta.json
        with open(os.path.join(root_folder, "meta.json"), "w") as f:
            json.dump(
                {
                    "system": self.system,
                    "symptom": self.symptom.value,
                    "stage": self.stage.value,
                    "version": self.version,
                    "version_id": self.version_id,
                },
                f,
            )

    @staticmethod
    def load(model_type, root_folder: str, allow_partial=False) -> "BugReport":

        symptom = None
        stage = None
        version = None
        version_id = None
        system = None

        assert allow_partial or os.path.exists(
            os.path.join(root_folder, "meta.json")
        ), "meta.json must exist or allow_partial is True where the oracle will be automatically generated"

        if os.path.exists(os.path.join(root_folder, "meta.json")):
            with open(os.path.join(root_folder, "meta.json"), "r") as f:
                meta = json.load(f)

            system = meta["system"]
            symptom = Symptom(meta["symptom"])
            stage = Stage(meta["stage"])
            version = meta["version"]
            version_id = meta["version_id"]

        testcase = TestCase.load(model_type, root_folder, allow_partial)

        log = None
        if os.path.exists(os.path.join(root_folder, BugReport.error_msg_name())):
            with open(os.path.join(root_folder, BugReport.error_msg_name()), "r") as f:
                log = f.read()

        return BugReport(testcase, symptom, stage, system, version, version_id, log)



import os

from nnsmith.backends import mk_factory, BackendFactory

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--backend", type=str, help="One of ort, trt, tvm, and xla", required=True
    )
    parser.add_argument("--device", type=str, default="cpu")
    parser.add_argument("--root", type=str, help="Path to the bug folder")
    args = parser.parse_args()

    fac = mk_factory(args.backend, args.device, optmax=True)

    crash_msg = set()

    for dir in os.listdir(args.root):
        path = os.path.join(args.root, dir)
        if not os.path.isdir(path):
            continue
        if not dir.startswith("bug-"):
            continue
        if dir.startswith("bug-torch") or dir.startswith("bug-omin"):
            continue

        try:
            onnx_path = os.path.join(path, "model.onnx")
            onnx_model = BackendFactory.get_onnx_proto(onnx_path)
            try:
                fac.make_backend(onnx_model)
            except Exception as e:
                crash_msg.add(str(e))
        except Exception as e:
            print(e)  # filter model-too-large kind of things.
            continue

    print(f"{args.root} got {len(crash_msg)} different crash messages:")
    err_msg_path = os.path.join(args.root, "crash_msg.txt")
    print(f"Writing unique crash messages to {err_msg_path}")
    if crash_msg:
        with open(err_msg_path, "w") as f:
            for msg in crash_msg:
                print(msg, file=f)
                print("$\n", file=f)  # splitter



import torch
from multipledispatch import dispatch

from nnsmith.abstract.op import *


@torch.jit.ignore
def numeric_valid(outputs) -> bool:
    with torch.no_grad():
        return all([torch.isfinite(out).all() for out in outputs])


# generalized loss fn
def smoothed_relu(x):
    if x.dtype == torch.float16:
        return torch.relu(x.float()).half()
    return torch.relu(x)


def loss_ge_zero(x):
    return smoothed_relu(-x)


def loss_le_zero(x):
    return smoothed_relu(x)


def loss_lt_zero(x):
    return loss_le(x, -1e-10)


def loss_gt_zero(x):
    return loss_ge(x, 1e-10)


def loss_ge(x, y):
    return loss_ge_zero(x - y)


def loss_le(x, y):
    return loss_le_zero(x - y)


def loss_gt(x, y):
    return loss_gt_zero(x - y)


def loss_lt(x, y):
    return loss_lt_zero(x - y)


# loss_fn:     backward


@dispatch(Div)
def loss_fn(op: Div):
    return lambda op, y: loss_gt_zero(torch.abs(y))


@dispatch(Pow)
def loss_fn(op: Pow):
    def torch_loss(a, b):
        # a >= 0 && b*log(a) <= 20
        l0 = loss_gt_zero(a)
        if torch.any(l0 > 0):
            return ("l0", l0)
        l1 = loss_le(
            b * torch.log(torch.maximum(a, torch.tensor(1e-40, dtype=a.dtype))), 40
        )
        return ("l1", l1)

    return torch_loss


@dispatch(Acos)
def loss_fn(op: Acos):
    return lambda x: loss_le(x.abs(), 1)


@dispatch(Sqrt)
def loss_fn(op: Sqrt):
    return lambda x: loss_ge(x, 0)


@dispatch(Asin)
def loss_fn(op: Asin):
    return lambda x: loss_le(x.abs(), 1)


@dispatch(Log2)
def loss_fn(op: Log2):
    return lambda x: loss_gt_zero(x)



import sys
import pandas as pd
import os

fuzz_report_rt = sys.argv[1]
df = pd.read_pickle(os.path.join(fuzz_report_rt, "profile.pkl"))
mstr = [f"{i}.onnx" for i in range(len(df))]
df["model_name"] = mstr
df["gen_t"] = df["gen_t"]
df1 = df[["gen_t", "model_name"]]
df1.to_csv(os.path.join(fuzz_report_rt, "gentime.csv"), index=False, header=False)



"""Just to figure out operators types and connections.
"""

from collections import Counter
import os
from multiprocessing import cpu_count, Process
import traceback
from typing import List, Tuple

from uuid import uuid4
import pandas as pd
import pickle

import matplotlib.pyplot as plt
from tqdm import tqdm
import numpy as np

from onnx_graph_analyzer import analyze_one_relay

SMALL_SIZE = 12
MEDIUM_SIZE = 16
BIGGER_SIZE = 18
SUPER_BIG = 22

plt.rcParams.update({"text.usetex": True})

plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
plt.rc("axes", titlesize=MEDIUM_SIZE)  # fontsize of the axes title
plt.rc("axes", labelsize=SUPER_BIG)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=BIGGER_SIZE)  # fontsize of the tick labels
plt.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc("legend", fontsize=SUPER_BIG)  # legend fontsize
plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title

__TMP_FOLDER__ = "param_analyzer_tmp"


def analyze_one(model_path, verbose=False) -> Tuple[str, Process]:
    def execute(model_path, target_path):
        try:
            res = analyze_one_relay(model_path, use_counter=True)
            if not os.path.exists(__TMP_FOLDER__):
                os.makedirs(__TMP_FOLDER__)
            with open(target_path, "wb") as f:
                pickle.dump(res, f)
            return target_path
        except:
            print("-------------> Skip model", model_path, "due to exception:")
            if verbose:
                traceback.print_exc()
            return None

    target_path = os.path.join(__TMP_FOLDER__, f"{uuid4()}.pkl")
    p = Process(target=execute, args=(model_path, target_path))
    p.start()
    return target_path, p


def analyze_folders(
    folders, cache_dir=None, force=False, n_limit=None, resume=False, write_freq=250
):
    res = []

    def write_once(res):
        if cache_dir is not None:
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)
            with open(os.path.join(cache_dir, __CACHE_FILE__), "wb") as fp:
                pickle.dump(res, fp)

    __CACHE_FILE__ = "onnx_param_cache.pkl"
    use_cache = (
        os.path.exists(os.path.join(cache_dir, __CACHE_FILE__))
        and cache_dir is not None
    )
    if resume:
        assert use_cache
    if use_cache and not force:
        print("===> {} already exists.".format(os.path.join(cache_dir, __CACHE_FILE__)))
        with open(os.path.join(cache_dir, __CACHE_FILE__), "rb") as fp:
            if resume:
                res = pickle.load(fp)
            else:
                return pickle.load(fp)

    times = []
    file_hubs = []
    least_time = None
    assert n_limit is None or len(n_limit) == len(folders)
    for i, folder in enumerate(folders):
        df = pd.read_csv(
            os.path.join(folder, "gentime.csv"), usecols=[0, 1], header=None
        )
        ts = df[0].to_numpy()

        times.append(ts)
        files = df[1].tolist()
        if n_limit is not None:
            files = files[: n_limit[i]]
        files = [os.path.join(folder, f) for f in files]
        file_hubs.append(files)

        time_cost = ts.sum()
        if least_time is None or time_cost < least_time:
            least_time = time_cost

    # only consider rows that < least_time
    for i, ts in enumerate(times):
        file_hubs[i] = file_hubs[i][: (ts < least_time).sum()]

    first_hub_resume = False
    if resume:
        new_file_hub = []
        if len(res) > 0:
            last_job_idx = len(res) - 1
            first_hub_done = res[last_job_idx]["nfiles"]
            if first_hub_done < len(file_hubs[last_job_idx]):
                new_file_hub.append(file_hubs[last_job_idx][first_hub_done:])
                first_hub_resume = True
        new_file_hub.extend(file_hubs[len(res) :])
        file_hubs = new_file_hub
        print(f"===> Resume from {len(file_hubs)} jobs.")

    def update_once(path):
        if os.path.exists(path):
            with open(path, "rb") as fp:
                new = pickle.load(fp)
                for op_name, cnt in new.items():
                    if op_name not in res[-1]["param_map"]:
                        res[-1]["param_map"][op_name] = Counter()
                    res[-1]["param_map"][op_name].update(cnt)
            os.remove(path)

    for files in file_hubs:
        if not first_hub_resume:
            res.append(
                {
                    "nfiles": 0,
                    "param_map": {},
                }
            )
        else:
            first_hub_resume = False

        n_pararllel = min(cpu_count() + 4, len(files))
        jobs: List[Tuple[str, Process]] = []

        def try_pop(timeout=None):
            if len(jobs) == 0:
                return False
            tar, p = jobs[0]
            p.join(timeout=timeout)
            if p.is_alive():
                return False
            if p.exitcode != 0:
                print(
                    "-------------> Skip model",
                    path,
                    "due to crash | exit code",
                    p.exitcode,
                )
            res[-1]["nfiles"] += 1
            pbar.update()
            update_once(tar)
            jobs.pop(0)
            return True

        with tqdm(total=len(files)) as pbar:
            for path in files:
                if len(jobs) < n_pararllel:
                    jobs.append(analyze_one(path, verbose=False))
                else:
                    try_pop()
                    while try_pop(timeout=0.1):
                        pass

                if pbar.n % write_freq == 0:
                    write_once(res)

            while try_pop():
                pass

    write_once(res)
    return res


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--folders", type=str, nargs="+", required=True)
    parser.add_argument("--tags", type=str, nargs="+", default=None)
    # should compare models within same generation duration.
    parser.add_argument("--nlim", type=int, nargs="+", default=None)
    parser.add_argument("--output", type=str, default="results")
    parser.add_argument("--force", action="store_true")
    parser.add_argument("--resume", action="store_true")
    parser.add_argument(
        "--ops", type=str, nargs="+", default=None, help="Pass operator names to show."
    )
    args = parser.parse_args()

    ops = args.ops
    if ops is None:
        ops = [
            "add",
            "nn.avg_pool2d",
            "nn.conv2d",
            "nn.dense",
            "strided_slice",
            "reshape",
        ]
    if args.tags is None:
        args.tags = [os.path.split(f)[-1].split("-")[0] for f in args.folders]
    else:
        assert len(args.tags) == len(args.folders)

    results = analyze_folders(
        args.folders,
        cache_dir=args.output,
        force=args.force,
        n_limit=args.nlim,
        resume=args.resume,
    )

    mutual_keys = set.intersection(*[set(r["param_map"].keys()) for r in results])

    # dyn -> dynamic op not considered
    # copy -> memory op not considered
    BLACK_LIST = ["dyn.", "copy"]
    # You at least to have 5 instances otherwise too trivial to show.
    NUM_THRESH = 5

    for k in list(mutual_keys):
        for bk in BLACK_LIST:
            if bk in k:
                mutual_keys -= set([k])

    same_keys = []
    # Don't plot if the item num is the same (meaningless to compare)
    for k in mutual_keys:
        if len(set([len(r["param_map"][k]) for r in results])) == 1:
            same_keys.append(k)

    for res in results:
        for k in list(mutual_keys):
            if len(res["param_map"][k]) < NUM_THRESH:
                mutual_keys -= set([k])

    mutual_keys -= set(same_keys)

    # only show representative ops
    #   merge redundancy;
    #      -> param-free element wise ops
    unary_pf_ew = [
        "abs",
        "acos",
        "cos",
        "asin",
        "sin",
        "ceil",
        "floor",
        "log",
        "logical_not",
        "atan",
        "tan",
        "erf",
        "negative",
        "nn.relu",
        "round",
        "sigmoid",
        "sqrt",
        "clip",
        "cast",
        "nn.prelu",
        "nn.leaky_relu",
    ]
    bin_pf_ew = [
        "add",
        "divide",
        "equal",
        "floor_mod",
        "greater",
        "multiply",
        "subtract",
        "less",
        "logical_and",
        "logical_or",
        "logical_xor",
        "maximum",
        "minimum",
        "power",
    ]
    simple_reduce_ops = ["sum", "min", "max", "mean"]
    UEW_KEY = "$\\bf unary$ elem.-wise"
    BEW_KEY = "$\\bf bin.$ elem.-wise"
    REDUCE_KEY = "common $\\bf reduce$"
    for res in results:
        res["param_map"][UEW_KEY] = Counter()
        for uk in unary_pf_ew:
            if uk in res["param_map"]:
                res["param_map"][UEW_KEY] += res["param_map"][uk]
        res["param_map"][BEW_KEY] = Counter()
        for bk in bin_pf_ew:
            if bk in res["param_map"]:
                res["param_map"][BEW_KEY] += res["param_map"][bk]
        res["param_map"][REDUCE_KEY] = Counter()
        for rk in simple_reduce_ops:
            if rk in res["param_map"]:
                res["param_map"][REDUCE_KEY] += res["param_map"][rk]

    mutual_keys -= set(unary_pf_ew)
    mutual_keys -= set(bin_pf_ew)
    mutual_keys -= set(simple_reduce_ops)

    mutual_keys.add(UEW_KEY)
    mutual_keys.add(BEW_KEY)
    mutual_keys.add(REDUCE_KEY)

    mutual_keys = sorted(list(mutual_keys))

    col_width = 3
    bar_width = (col_width * 0.5) / len(results)
    base_x = np.arange(len(mutual_keys))

    vals = []
    for single_res in results:
        param_map = single_res["param_map"]
        v = []
        for k in mutual_keys:
            v.append(len(param_map[k]))
        vals.append(v)

    vals = np.array(vals)

    legends = []
    HATCHES = ["x", None]
    COLORS = ["lightblue", "dodgerblue"]

    fig, ax = plt.subplots(1, 1, constrained_layout=True, figsize=(13, 5))

    print(mutual_keys)

    for idx, (tag, single_res) in enumerate(zip(args.tags, results)):
        param_map = single_res["param_map"]
        print(f"{single_res['nfiles']=}")
        padding = int((10 - len(tag)) * 1.2)
        blank = r"\ "
        legends.append(
            f'{tag + "".join([blank for _ in range(padding)])} (total: {vals[idx].sum()})'
        )

    normalized = vals.max(axis=0) / vals.min(axis=0)
    indices = np.argsort(normalized)

    for idx, (tag, single_res) in enumerate(zip(args.tags, results)):
        pv = vals[idx] / np.min(vals, axis=0)

        x_pos = base_x  # - 0.5 * col_width + (idx + 0.5) * bar_width
        ax.bar(
            x_pos,
            pv[indices],
            width=bar_width,
            label=tag,
            color=COLORS[idx],
            align="center",
            hatch=HATCHES[idx],
            edgecolor="k",
            linewidth=3,
        )

    for x, v in zip(base_x, normalized[indices]):
        ax.text(
            x - bar_width * 0.54,
            v + 0.13,
            f"{v:.1f}$\\times$",
            fontweight="bold",
            fontsize=MEDIUM_SIZE,
            fontname="Times New Roman",
        )

    ax.get_yaxis().set_ticks([])
    plt.legend(legends)

    xticks = []
    for idx in indices:
        k = mutual_keys[idx]
        xticks.append(k if k.startswith("$") else k.split(".")[-1])
    plt.xticks(base_x, xticks, rotation=35, ha="right", rotation_mode="anchor")
    plt.xlim([base_x[0] - 0.5, base_x[-1] + 0.5])
    plt.ylim([0, normalized.max() + 0.8])
    # plt.gcf().subplots_adjust(left=base_x[0] - 1, right=base_x[-1] + 1)
    # fig.subplots_adjust(left=base_x[0] - 1, right=base_x[-1] + 1)
    # plt.xlabel('# Operators in Models w/ Vulnerable Op.', fontweight='bold')
    plt.ylabel("Unique Operator Instances", fontweight="bold")
    plt.savefig(os.path.join(args.output, "param_diff.pdf"))
    plt.savefig(os.path.join(args.output, "param_diff.png"))



import random
from typing import List, Tuple, Union

from nnsmith.abstract.arith import *
from nnsmith.abstract.dtype import DType
from nnsmith.abstract.op import (
    BcastBinaryOp,
    ElementWiseUnaryOp,
    MatMul,
    UnaryOpBase,
    mark_materialize,
    rank_from,
)
from nnsmith.abstract.tensor import AbsTensor
from nnsmith.error import ConstraintCheck


@mark_materialize("tensorflow")
class Dense(UnaryOpBase):
    in_dtypes = [(DType.float32,), (DType.float64,)]
    out_dtypes = [(DType.float32,), (DType.float64,)]

    def __init__(self, ifeat: Union[int, z3.ExprRef], ofeat: Union[int, z3.ExprRef]):
        super().__init__()
        self.ifeat = ifeat
        self.ofeat = ofeat
        self.inp_ranks = [
            rank_from(2)
        ]  # NOTE: tensorflow Dense layer requires an input with batch as its first axis
        # at least one dim. cannot be zranks_all()
        self.out_ranks = [rank_from(2)]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        assert len(input_shapes) == 1, "Linear only takes one input, but got {}".format(
            len(input_shapes)
        )
        return [
            AbsTensor(
                shape=[*input_shapes[0].shape[:-1], self.ofeat],
                dtype=input_shapes[0].dtype,
            )
        ]

    def requires(self, input_shapes: List[AbsTensor]) -> List[z3.ExprRef]:
        ConstraintCheck.true(input_shapes[0].ndims >= 2)
        return [
            nnsmith_ge(self.ifeat, 1),
            nnsmith_ge(self.ofeat, 1),
            nnsmith_eq(input_shapes[0].shape[-1], self.ifeat),
        ]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(out_abs_tensor[0].ndims, out_abs_tensor[0].dtype)]


@mark_materialize("tensorflow")
class SquaredDifference(BcastBinaryOp):
    pass


@mark_materialize("tensorflow")
class LocalRespNorm(ElementWiseUnaryOp):
    # https://www.tensorflow.org/api_docs/python/tf/raw_ops/LRN
    in_dtypes = [(DType.float16,), (DType.float32,)]
    out_dtypes = [(DType.float16,), (DType.float32,)]

    def __init__(
        self,
        depth_radius: Union[int, z3.ExprRef],
    ):
        super().__init__()
        self.depth_radius = depth_radius
        self.extra_attrs["bias"] = random.uniform(0.01, 100)
        self.extra_attrs["alpha"] = random.uniform(0.01, 100)
        # cuDNN requires beta >= 0.01
        self.extra_attrs["beta"] = random.uniform(0.011, 1)

        self.inp_ranks = [(4,)]
        self.out_ranks = [(4,)]

    def requires(self, input_shapes: List[AbsTensor]) -> List[z3.ExprRef]:
        SanityCheck.eq(len(input_shapes), 1)
        input_shape = input_shapes[0]
        cons = []
        # cuDNN requires depth_radius in [1, 7]
        cons.append(nnsmith_ge(self.depth_radius, 1))
        cons.append(nnsmith_le(self.depth_radius, 7))
        cons.append(
            nnsmith_le(
                self.depth_radius, nnsmith_div(nnsmith_sub(input_shape.shape[3], 1), 2)
            )
        )  # depth_radius <= (input_shape[3] - 1) / 2
        return cons


class NHWCConv2d(UnaryOpBase):
    in_dtypes = [(DType.float16,), (DType.float32,), (DType.float64,)]
    out_dtypes = [(DType.float16,), (DType.float32,), (DType.float64,)]

    def __init__(
        self,
        in_channels: Union[int, z3.ExprRef],
        out_channels: Union[int, z3.ExprRef],
        kernel_h_size: Union[int, z3.ExprRef],
        kernel_w_size: Union[int, z3.ExprRef],
        stride: Union[int, z3.ExprRef],
        dilation_h: Union[int, z3.ExprRef],
        dilation_w: Union[int, z3.ExprRef],
        padding: str,
    ):
        """See https://www.tensorflow.org/api_docs/python/tf/keras/layers/Conv2D"""
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_h_size = kernel_h_size
        self.kernel_w_size = kernel_w_size
        self.stride = stride
        self.dilation_h = dilation_h
        self.dilation_w = dilation_w
        SanityCheck.true(padding in ["valid", "same"])
        self.extra_attrs["padding"] = padding
        self.inp_ranks = [(4,)]
        self.out_ranks = [(4,)]

    def __str__(self) -> str:
        return (
            self.name()
            + f" (kernel={(self.kernel_h_size,self.kernel_w_size)}, "
            + f"stride={self.stride}, "
            + f"dilation={(self.dilation_h,self.dilation_w)}, "
            + f"ochannels={self.out_channels})"
        )

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        # https://www.tensorflow.org/api_docs/python/tf/nn#notes_on_padding_2
        ni, hi, wi, _ = input_shapes[0].shape
        no = ni
        co = self.out_channels
        mimic_kh = self.kernel_h_size + (self.dilation_h - 1) * (self.kernel_h_size - 1)
        mimic_kw = self.kernel_w_size + (self.dilation_w - 1) * (self.kernel_w_size - 1)
        if self.extra_attrs["padding"] == "valid":
            ho = nnsmith_div(hi - mimic_kh, self.stride) + 1
            wo = nnsmith_div(wi - mimic_kw, self.stride) + 1
        elif self.extra_attrs["padding"] == "same":
            ho = nnsmith_div(hi + self.stride - 1, self.stride)
            wo = nnsmith_div(wi + self.stride - 1, self.stride)
        return [AbsTensor(shape=[no, ho, wo, co], dtype=input_shapes[0].dtype)]

    def requires(self, input_shapes):
        _, hi, wi, ci = input_shapes[0].shape
        mimic_kh = self.kernel_h_size + (self.dilation_h - 1) * (self.kernel_h_size - 1)
        mimic_kw = self.kernel_w_size + (self.dilation_w - 1) * (self.kernel_w_size - 1)

        cons = [
            nnsmith_eq(self.in_channels, ci),
            nnsmith_ge(self.out_channels, 1),
            nnsmith_ge(self.dilation_h, 1),
            nnsmith_ge(self.dilation_w, 1),
            nnsmith_ge(mimic_kh, 1),
            nnsmith_ge(mimic_kw, 1),
            nnsmith_ge(self.stride, 1),
        ]

        if self.extra_attrs["padding"] == "valid":
            cons.append(nnsmith_le(mimic_kh, hi))
            cons.append(nnsmith_le(mimic_kw, wi))

        # The following constraint is from TensorFlow tracing:
        # `strides > 1` not supported in conjunction with `dilation_rate > 1`
        cons.append(  # 0 == (stride - 1) * (max(dh, dw) - 1)
            nnsmith_or(
                nnsmith_eq(self.stride, 1),
                nnsmith_and(
                    nnsmith_eq(self.dilation_h, 1),
                    nnsmith_eq(self.dilation_w, 1),
                ),
            )
        )
        return cons

    def flops(self, input_shapes):
        return nnsmith_mul(
            nnsmith_mul(
                nnsmith_mul(
                    self.type_transfer(input_shapes)[0].nelement(), self.in_channels
                ),
                self.kernel_h_size,
            ),
            self.kernel_w_size,
        )

    def n_floats(self, input_shapes):
        # FIXME: maybe need to take dilation into account?
        padding = 0
        padded_data = AbsTensor(input_shapes[0].shape, dtype=input_shapes[0].dtype)
        padded_data.shape[2] = nnsmith_add(
            padded_data.shape[2], nnsmith_mul(2, padding)
        )
        padded_data.shape[3] = nnsmith_add(
            padded_data.shape[3], nnsmith_mul(2, padding)
        )
        w = AbsTensor(
            [
                self.out_channels,
                self.in_channels,
                self.kernel_h_size,
                self.kernel_w_size,
            ],
            dtype=input_shapes[0].dtype,
        )
        outs = super().n_floats(input_shapes)
        return nnsmith_add(nnsmith_add(w.nelement(), padded_data.nelement()), outs)

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(4, out_abs_tensor[0].dtype)]


@mark_materialize("tensorflow")
class NHWCConv2dValidPad(NHWCConv2d):
    def __init__(
        self,
        in_channels: Union[int, z3.ExprRef],
        out_channels: Union[int, z3.ExprRef],
        kernel_h_size: Union[int, z3.ExprRef],
        kernel_w_size: Union[int, z3.ExprRef],
        stride: Union[int, z3.ExprRef],
        dilation_h: Union[int, z3.ExprRef],
        dilation_w: Union[int, z3.ExprRef],
    ):
        super().__init__(
            in_channels,
            out_channels,
            kernel_h_size,
            kernel_w_size,
            stride,
            dilation_h,
            dilation_w,
            "valid",
        )


@mark_materialize("tensorflow")
class NHWCConv2dSamePad(NHWCConv2d):
    def __init__(
        self,
        in_channels: Union[int, z3.ExprRef],
        out_channels: Union[int, z3.ExprRef],
        kernel_h_size: Union[int, z3.ExprRef],
        kernel_w_size: Union[int, z3.ExprRef],
        stride: Union[int, z3.ExprRef],
        dilation_h: Union[int, z3.ExprRef],
        dilation_w: Union[int, z3.ExprRef],
    ):
        super().__init__(
            in_channels,
            out_channels,
            kernel_h_size,
            kernel_w_size,
            stride,
            dilation_h,
            dilation_w,
            "same",
        )


@mark_materialize("tensorflow")
class TFMatMul(MatMul):
    def __init__(self):
        super().__init__()
        self.inp_ranks = [(2, 3), (2, 3)]
        self.out_ranks = [(2, 3)]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        if out_abs_tensor[0].ndims == 2:
            return [
                (2, out_abs_tensor[0].dtype),
                (2, out_abs_tensor[0].dtype),
            ]
        # at least one of them is 3
        ranks = [3, random.choice([2, 3])]
        random.shuffle(ranks)
        return [
            (ranks[0], out_abs_tensor[0].dtype),
            (ranks[1], out_abs_tensor[0].dtype),
        ]



"""
Given the directory containing all tests. We replay the test execution and record coverage in LLVM profraw format.
The intermediate tests can be saved using fuzz.save_test={{DIR_TO_SAVE}}.
"""
import multiprocessing as mp
import os
import subprocess

from nnsmith.util import mkdir


def batched(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]


def model_exec(test_paths, model_type, backend_type, backend_target, profraw_path):
    model_paths = []
    for test_path in test_paths:
        for file in os.listdir(test_path):
            if file.startswith("model"):
                model_paths.append(os.path.join(test_path, file))
                break

    arguments = [
        "python3",
        "nnsmith/cli/model_exec.py",
        "model.type=" + model_type,
        "backend.type=" + backend_type,
        "backend.target=" + backend_target,
        f"model.path={model_paths}",
    ]

    copied_env = os.environ.copy()
    copied_env["LLVM_PROFILE_FILE"] = profraw_path

    p = subprocess.Popen(
        arguments,  # Show all output
        env=copied_env,
    )
    p.communicate()
    exit_code = p.returncode

    if exit_code != 0:
        print(
            f"==> model_exec crashed when generating {profraw_path}! => EXIT CODE {exit_code}"
        )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--root", type=str, required=True, help="Folder to all the tests."
    )
    parser.add_argument("--batch_size", type=int, default=100, help="")
    parser.add_argument("--model_type", type=str, required=True, help="Model type.")
    parser.add_argument("--backend_type", type=str, required=True, help="Backend type.")
    parser.add_argument(
        "--backend_target", type=str, required=True, help="Say `cpu` or `cuda`."
    )
    parser.add_argument(
        "--parallel", type=int, default=8, help="Number of process for execution."
    )

    args = parser.parse_args()

    time2path = {}
    for dir in os.listdir(args.root):
        if dir != "coverage":
            time2path[float(dir)] = os.path.join(args.root, dir)

    time_stamps = sorted(time2path.keys())
    batches = list(batched(time_stamps, args.batch_size))

    print(f"=> Number of batches: {len(batches)} of size {args.batch_size}")

    cov_save = os.path.join(args.root, "coverage")

    mkdir(cov_save)

    def batch_exec(batch):
        batch_paths = [time2path[time] for time in batch]
        profraw_path = os.path.join(cov_save, f"{max(batch)}.profraw")
        model_exec(
            batch_paths,
            args.model_type,
            args.backend_type,
            args.backend_target,
            profraw_path,
        )

    with mp.Pool(processes=args.parallel) as pool:
        pool.map(batch_exec, batches)



import os
import random
import warnings
from io import BytesIO
from os import PathLike
from typing import Dict, Generic, List, Optional, Tuple, Type, TypeVar, Union

import onnx
import onnx.checker
import onnx.helper
import torch
import torch.onnx
from onnx.external_data_helper import load_external_data_for_model
from onnx.tools import update_model_dims

from nnsmith.abstract.dtype import DTYPE_GEN_COMPLEX, DType
from nnsmith.abstract.op import AbsTensor
from nnsmith.gir import GraphIR
from nnsmith.macro import onnx2external_data_dir
from nnsmith.materialize.torch import (
    SymbolNet,
    TorchModel,
    TorchModelCPU,
    TorchModelCUDA,
)


def create_deadcode_onnx(onnx_model: onnx.ModelProto, name_mask) -> onnx.ModelProto:
    graph_def = onnx.helper.make_graph(
        nodes=onnx_model.graph.node,  # nodes
        name=onnx_model.graph.name,  # name
        inputs=onnx_model.graph.input,  # inputs
        outputs=[o for o in onnx_model.graph.output if o.name in name_mask],  # outputs
        initializer=onnx_model.graph.initializer,
    )

    model_def = onnx.helper.make_model(
        graph=graph_def, producer_name="nnsmith.deadcode"
    )

    onnx.checker.check_model(model_def, full_check=True)
    return model_def


def torch2onnx(
    model: SymbolNet,
    exportable: Union[str, BytesIO],
    verbose=False,
    dummy_inputs=None,
    do_constant_folding=False,
) -> None:
    """Convert PyTorch model to ONNX format."""
    proxy_enabled = model.proxy_enabled
    if proxy_enabled:
        model.disable_proxy_grad()

    # Dummy inputs
    if dummy_inputs is None:
        dummy_inputs = [
            torch.ones(size=svar.shape)
            .to(model.device)
            .uniform_(1, 2)
            .to(dtype=svar.dtype.torch())
            for _, svar in model.input_like.items()
        ]

    input_names = list(model.input_like.keys())

    with torch.no_grad():
        with warnings.catch_warnings():
            warnings.simplefilter(
                "default" if verbose else "ignore", category=torch.jit.TracerWarning
            )
            warnings.simplefilter(
                "default" if verbose else "ignore", category=UserWarning, append=True
            )
            model.eval()
            # TODO: How to remove the annoying "Diagnostic Run ..." message?
            torch.onnx.export(
                model,
                tuple(dummy_inputs),
                exportable,
                input_names=input_names,
                output_names=list(model.output_like.keys()),
                verbose=verbose,
                do_constant_folding=do_constant_folding,
                opset_version=14,
            )

    if proxy_enabled:  # Re-enable proxy grad
        model.enable_proxy_grad()


def dtype_from_onnx(onnx_dtype: onnx.TensorProto.DataType) -> DType:
    """Return dtype from ONNX data type."""
    return {
        onnx.TensorProto.DataType.FLOAT: DType.float32,
        onnx.TensorProto.DataType.DOUBLE: DType.float64,
        onnx.TensorProto.DataType.UINT8: DType.uint8,
        onnx.TensorProto.DataType.INT8: DType.int8,
        onnx.TensorProto.DataType.INT16: DType.int16,
        onnx.TensorProto.DataType.INT32: DType.int32,
        onnx.TensorProto.DataType.INT64: DType.int64,
        onnx.TensorProto.DataType.BOOL: DType.bool,
        onnx.TensorProto.DataType.COMPLEX64: DType.complex64,
        onnx.TensorProto.DataType.COMPLEX128: DType.complex128,
        onnx.TensorProto.DataType.FLOAT16: DType.float16,
        # onnx.TensorProto.DataType.UINT8: DType.uint8, // unsigned type not supported
        # onnx.TensorProto.DataType.UINT16: DType.uint16,
        # onnx.TensorProto.DataType.UINT32: DType.uint32,
        # onnx.TensorProto.DataType.UINT64: DType.uint64,
        # onnx.TensorProto.DataType.STRING: DType.string, // strings not supported
        # onnx.TensorProto.DataType.BFLOAT16: DType.bfloat16, // bfloat16 not supported
    }[onnx_dtype]


def analyze_onnx_io(
    model: onnx.ModelProto,
) -> Tuple[Dict[str, AbsTensor], Dict[str, AbsTensor]]:
    """Return input and output types in dictionaries."""
    input_types = {}
    output_types = {}
    for input in model.graph.input:
        name = input.name
        dtype = dtype_from_onnx(input.type.tensor_type.elem_type)
        shape = [dim.dim_value for dim in input.type.tensor_type.shape.dim]
        # assert all shapes are positive integers
        assert all(
            isinstance(dim, int) and dim >= 0 for dim in shape
        ), f"Fixed shape needed, but got {shape} for input {name}"
        input_types[name] = AbsTensor(shape=shape, dtype=dtype)
    for output in model.graph.output:
        name = output.name
        dtype = dtype_from_onnx(output.type.tensor_type.elem_type)
        shape = [dim.dim_value for dim in output.type.tensor_type.shape.dim]
        assert all(
            isinstance(dim, int) and dim >= 0 for dim in shape
        ), f"Fixed shape needed, but got {shape} for output {name}"
        output_types[name] = AbsTensor(shape=shape, dtype=dtype)
    return input_types, output_types


def get_onnx_proto(model: Union[onnx.ModelProto, str]) -> onnx.ModelProto:
    if isinstance(model, onnx.ModelProto):
        return model
    else:
        external_data_dir = onnx2external_data_dir(model)
        if os.path.exists(external_data_dir):
            onnx_model = onnx.load(model, load_external_data=False)
            load_external_data_for_model(onnx_model, external_data_dir)
        else:
            onnx_model = onnx.load(model)
        return onnx_model


class ONNXModel(TorchModel):
    PTType: Type[TorchModel] = None

    @classmethod
    def device(cls) -> torch.device:
        return cls.PTType.device()

    def __init__(self, with_torch=True):
        """Initialize a ONNXModel.

        Args:
            with_torch (bool, optional): Whether to load/dump torch related files. Defaults to True.
        """
        super().__init__()
        self.with_torch: bool = with_torch
        self.masked_output_like: Optional[Dict[str, AbsTensor]] = None
        self.full_output_like: Optional[Dict[str, AbsTensor]] = None
        self.full_input_like: Optional[Dict[str, AbsTensor]] = None
        self.onnx_model = None

    @property
    def version(self) -> str:
        return f"onnx{onnx.__version__}-torch{torch.__version__}"

    def _mask_outputs(self) -> Dict[str, AbsTensor]:
        assert (
            self.torch_model is not None
        ), "Create a concrete model before masking outputs."

        # random mask that has mask_n True, otherwise False.
        after_mask = {
            k: v for k, v in self.full_output_like.items() if random.random() < 0.5
        }
        if len(after_mask) == 0:
            # at least one output must be returned.
            only_key = random.sample(self.full_output_like.keys(), 1)[0]
            after_mask = {only_key: self.full_output_like[only_key]}

        return after_mask

    @staticmethod
    def _dce_prob() -> float:  # \in [0, 1]
        dce_prob = 0.0
        dce_env = os.getenv("NNSMITH_ONNX_DCE")
        if dce_env is not None:
            if not dce_env.replace(".", "", 1).isdigit() or not 0 < float(dce_env) < 1:
                raise ValueError(f"NNSMITH_ONNX_DCE must be [0, 1], but got {dce_env}")
            dce_prob = float(dce_env)
        return dce_prob

    @classmethod
    def from_gir(cls: Type["ONNXModel"], gir: GraphIR, **kwargs) -> "ONNXModel":
        ret = cls()
        ret.torch_model = cls.PTType.from_gir(gir, **kwargs).torch_model
        ret.full_input_like = ret.torch_model.input_like
        ret.full_output_like = ret.torch_model.output_like
        ret.masked_output_like = ret.full_output_like

        if random.random() < cls._dce_prob():
            ret.masked_output_like = ret._mask_outputs()

        return ret

    def refine_weights(self) -> None:
        self.PTType.refine_weights(self)
        # weights are set. let's save the model.
        self.onnx_model = self.get_onnx_from_torch()
        if set(self.masked_output_like.keys()) != set(self.full_output_like):
            self.onnx_model = create_deadcode_onnx(
                self.onnx_model, self.masked_output_like.keys()
            )

    @property
    def input_like(self) -> Dict[str, AbsTensor]:
        return self.full_input_like

    @property
    def output_like(self) -> Dict[str, AbsTensor]:
        return self.masked_output_like

    def dump(self, path: PathLike) -> None:
        if self.with_torch:
            self.PTType.dump(
                self, path.replace(self.name_suffix(), self.PTType.name_suffix())
            )
            if self.onnx_model is None:
                self.onnx_model = self.get_onnx_from_torch()

        onnx.checker.check_model(self.onnx_model, full_check=True)
        onnx.save(self.onnx_model, path)

    @classmethod
    def load(cls, path: PathLike) -> "ONNXModel":
        ret = cls()
        ret.onnx_model = onnx.load(path)

        torch_path = path.replace(cls.name_suffix(), cls.PTType.name_suffix())

        ret.with_torch = False
        full_input_like, full_output_like = analyze_onnx_io(ret.onnx_model)
        ret.full_input_like = full_input_like
        ret.full_output_like = full_output_like
        ret.masked_output_like = ret.full_output_like

        # FIXME: missing key(s) in state_dict: "mlist.0.data", "mlist.1.data".
        if os.path.exists(torch_path):
            ret.with_torch = True
            ret.torch_model = cls.PTType.load(torch_path)
            ret.full_input_like = ret.torch_model.input_like
            ret.full_output_like = ret.torch_model.output_like

        return ret

    @property
    def native_model(self):
        if self.with_torch and self.onnx_model is None:
            self.onnx_model = self.get_onnx_from_torch()
        return self.onnx_model

    def get_onnx_from_torch(self) -> onnx.ModelProto:
        f = BytesIO()
        torch2onnx(self.torch_model, f)
        onnx_model = onnx.load_model_from_string(f.getvalue())
        # freeze input and output shapes.
        onnx_model = update_model_dims.update_inputs_outputs_dims(
            onnx_model,
            {k: v.shape for k, v in self.torch_model.input_like.items()},
            {k: v.shape for k, v in self.torch_model.output_like.items()},
        )
        onnx.checker.check_model(onnx_model, full_check=True)
        return onnx_model

    @staticmethod
    def name_suffix() -> str:
        return ".onnx"

    @classmethod
    def skip_dtypes(cls) -> List[DType]:
        return DTYPE_GEN_COMPLEX


class ONNXModelCPU(ONNXModel):
    PTType = TorchModelCPU


class ONNXModelCUDA(ONNXModel):
    PTType = TorchModelCUDA



"""
This is basically a simplied implementation of "Graph-Based Fuzz Testing for Deep Learning Inference Engine".
Here's a few facts:
1. They are not open-source.
2. We only implement how they generate a valid graph (basically ensure its capability).
3. (point 2 cont.) We don't care their complicated heuristics and corpus.
4. The key explanantion is in Section III.G "Shapes & parameters Calculator."
    4.1: In additoin to LEMON, they support non-unary operators, ensuring the shape match by:
        a. slicing;
        b. padding (keeping layers element-wise);
        c. unsqueezing; (we added this to make GraphFuzz even better)
    4.2: Operators with padding to try to ensure everything is element-wise.
        NOTE: Through the examples, they basically assume input rank is 4 (NCHW).

Thus we now conclude the search space:
- operators:
    - input/output rank allows: 4 or -1;
    - element-wise opeartors;
    - exceptions:
        pooling use padding to make it elementwise;
        convolution use padding to make it elementwise;
    - data type casting.
    - parameters shall be randomly genenrated;
- input:
    - single input; (they said MNN cannot support multi-input)
    - input rank: 4;
"""

import random
from typing import List
import warnings
import os
import time

import numpy as np
import torch
from tqdm import tqdm

from nnsmith.util import mkdir
from nnsmith.abstract.op import (
    ALL_OP_TYPES,
    AbsOpBase,
    Softmax,
    BatchNorm2d,
    Concat,
    Input,
    Constant,
    NCHWConv2d,
    Pool2d,
    DType,
    Div,
    ElementWiseUnaryOp,
    BcastBinaryOp,
)
from nnsmith.dtype_test import rewrite_op_dtype


class GraphFuzzNet(torch.nn.Module):
    def __init__(self, op_list: List[AbsOpBase]) -> None:
        super().__init__()
        self.op_list = op_list
        self.torch_inst = []
        self.mlist = torch.nn.ModuleList()
        for op in op_list:
            inst = op.torch()
            self.torch_inst.append(inst)
            if isinstance(inst, torch.nn.Module):
                self.mlist.append(inst)

    def forward(self, x):
        available_tensors = [x]
        shape_checker = x.shape
        for idx, op in enumerate(self.op_list):
            n_inp = len(op.inp_ranks)
            n_out = len(op.out_ranks)
            selected = [random.choice(available_tensors) for _ in range(n_inp)]
            dtype_match = False
            for its in op.in_dtypes:
                n_match = 0
                for expected, have in zip(its, selected):
                    if str(expected) != str(have.dtype):
                        break
                    n_match += 1
                if n_match == n_inp:
                    dtype_match = True
                    break

            # dtype match
            if not dtype_match:
                type_req = random.choice(op.in_dtypes)
                selected = [
                    tensor.type(DType.torch(t)) for tensor, t in zip(selected, type_req)
                ]

            # rank match
            for i_st, st in enumerate(selected):
                while len(selected[i_st].shape) > len(shape_checker):
                    selected[i_st] = torch.select(selected[i_st], dim=0, index=0)

                while len(selected[i_st].shape) < len(shape_checker):
                    selected[i_st] = selected[i_st].unsqueeze(0)

            # shape match
            for i_st, st in enumerate(selected):
                assert len(st.shape) == len(shape_checker)
                if not st.shape == shape_checker:
                    pads = []
                    need2pad = False
                    need2slice = False
                    for i in range(len(st.shape)):
                        if st.shape[i] < shape_checker[i]:
                            pads.append(shape_checker[i] - st.shape[i])
                            need2pad = True
                        elif st.shape[i] > shape_checker[i]:
                            pads.append(0)
                            need2slice = True
                        else:
                            pads.append(0)

                    # padding if too small
                    if need2pad:
                        padding = []
                        for p in pads:
                            padding.append(0)
                            padding.append(p)
                        selected[i_st] = torch.nn.functional.pad(
                            selected[i_st], tuple(padding), mode="constant", value=1
                        )

                    # slicing if too large
                    if need2slice:
                        # must be rank of 4 now.
                        selected[i_st] = selected[i_st][
                            : shape_checker[0],
                            : shape_checker[1],
                            : shape_checker[2],
                            : shape_checker[3],
                        ]

            if isinstance(op, Div):
                # avoid float-point errors.
                selected[1] = selected[1].abs() + 0.1

            outs = self.torch_inst[idx](*selected)

            if n_out == 1:
                available_tensors.append(outs)
            else:
                available_tensors.extend(outs)

        return available_tensors[-1]


class GraphFuzz:
    @staticmethod
    def get_available_op_ts(try_all=False):
        # Using try all, we allow GraphFuzz to generate all possible operators.
        # even though most of it gonna fail.
        available_op_ts = []
        for op_t in ALL_OP_TYPES:
            if op_t is Input or op_t is Constant:
                continue
            if not try_all:
                if issubclass(op_t, ElementWiseUnaryOp):
                    available_op_ts.append(op_t)
                elif op_t is NCHWConv2d:
                    available_op_ts.append(op_t)
                elif issubclass(op_t, Concat):
                    available_op_ts.append(op_t)
                elif issubclass(op_t, Pool2d):
                    available_op_ts.append(op_t)
                elif issubclass(op_t, BcastBinaryOp):
                    available_op_ts.append(op_t)
                continue
            available_op_ts.append(op_t)
        return available_op_ts

    def get_pool2d_params(self, input_shape):
        """To make input output element-wise.
        Pool   (kw, kh, s, pad)
        Input  (iN, iC, iH, iW)
        Output (oN, oC, oH, oW)
        -------------------------------------
        oN = iN
        oC = iC
        oH = iH = (iH - kh + 2 * pad) / s + 1
        oW = iW = (iW - kw + 2 * pad) / s + 1
        0 <= pad <= (kh - 1) / 2 # The condition in the paper for Conv2d is not applicable for pooling.
        0 <= pad <= (kw - 1) / 2
        1 <= s <= Max_sH (let's say it's 4)
        1 <= s <= Max_sW (let's say it's 4)
        """
        assert len(input_shape) == 4
        iN, iC, iH, iW = input_shape
        __MAX_TRY__ = 256
        kh = None
        kw = None
        s = None
        pad = None
        for _ in range(__MAX_TRY__):
            kw = random.randint(1, iW)
            kh = random.randint(1, iH)
            s = random.randint(1, 4)
            pad = random.randint(0, (kh - 1) // 2)
            if (iH - kh + 2 * pad) / s + 1 == iH and (iW - kw + 2 * pad) / s + 1 == iW:
                return kh, kw, s, pad
        return [1, 1, 1, 0]  # simple fallback

    def get_conv2d_params(self, input_shape):
        assert len(input_shape) == 4
        iN, iC, iH, iW = input_shape
        __MAX_TRY__ = 256
        kh = None
        kw = None
        s = None
        pad = None
        for _ in range(__MAX_TRY__):
            kw = random.randint(1, iW)
            kh = random.randint(1, iH)
            s = random.randint(1, 4)
            pad = random.randint(0, kh - 1)
            if (iH - kh + 2 * pad) / s + 1 == iH and (iW - kw + 2 * pad) / s + 1 == iW:
                return iC, iC, kh, kw, s, pad
        return [iC, iC, 1, 1, 1, 0]  # simple fallback

    def __init__(
        self, approx_nop=10, dim_limit=[5, 5, 224, 224], try_all=False
    ) -> None:
        self.available_op_ts = self.get_available_op_ts(try_all=try_all)
        self.base_op_n = approx_nop
        self.dim_limit = dim_limit

    def create_random_schedule(self):
        """
        basically return a list of ops.
        """
        input_shape = []
        for ls in self.dim_limit:
            input_shape.append(random.randint(1, ls))

        ops = []
        op_ts = [random.choice(self.available_op_ts) for _ in range(self.base_op_n)]
        for op_t in op_ts:
            if op_t is NCHWConv2d:
                ic, oc, kw, kh, s, pad = self.get_conv2d_params(input_shape)
                ops.append(op_t(ic, oc, kh, kw, s, pad))
            elif issubclass(op_t, Pool2d):
                kw, kh, s, pad = self.get_pool2d_params(input_shape)
                ops.append(op_t(kh, kw, s, pad))
            elif op_t is Softmax:
                ops.append(op_t(random.randint(1, len(input_shape) - 1)))
            elif op_t is BatchNorm2d:
                ops.append(op_t(input_shape[1]))
            elif issubclass(op_t, Concat):
                op = op_t()
                op.extra_attrs["axis"] = random.randint(0, 3)
                ops.append(op)
            else:
                ops.append(
                    op_t(
                        *[
                            random.randint(0, 10)
                            for _ in range(op_t.get_num_var_param())
                        ]
                    )
                )

        return ops, input_shape

    def run_once(self, save_path=None):
        ops, ishape = self.create_random_schedule()
        model = GraphFuzzNet(ops)
        with torch.no_grad():
            torch.onnx.export(model, torch.randn(ishape), save_path, opset_version=14)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--approx_nop", type=int, default=10)
    parser.add_argument("--dim_limit", type=int, nargs="+", default=[5, 5, 224, 224])
    parser.add_argument("--time_budget", type=int, default=60 * 60 * 4)
    parser.add_argument("--onnx_dir", type=str, default=None)
    parser.add_argument("--ort_cache", type=str, default=None)
    parser.add_argument("--seed", type=int, default=233)
    parser.add_argument("--try_all", action="store_true")
    args = parser.parse_args()

    print(f"Using seed {args.seed}")
    random.seed(args.seed)
    torch.manual_seed(args.seed)
    np.random.seed(args.seed)

    mkdir(args.onnx_dir)

    if args.ort_cache:
        if not os.path.exists(args.ort_cache):
            print(f"Please first generate cache! (mkdir config first)")
            print(f"python nnsmith/dtype_test.py --cache {args.ort_cache}")
            exit(1)
        # must pre run this. otherwise using ort will slow down generation.
        rewrite_op_dtype(ALL_OP_TYPES, factory=None, cache=args.ort_cache)

    gf = GraphFuzz(
        approx_nop=args.approx_nop, dim_limit=args.dim_limit, try_all=args.try_all
    )

    # FORMAT: {generation time cost in seconds}, {model relative path}
    # MUST RANK by GENERATION ORDER.
    config_file = open(os.path.join(args.onnx_dir, "gentime.csv"), "w")

    start_time = time.time()
    gen_cnt = 0
    valid_cnt = 0

    with tqdm(total=args.time_budget) as pbar:
        while time.time() - start_time < args.time_budget:
            seed = random.getrandbits(32)
            to_name = f"{valid_cnt}.onnx"

            tstart = time.time()
            try:
                with warnings.catch_warnings():  # just shutup.
                    warnings.simplefilter("ignore")
                    gf.run_once(save_path=os.path.join(args.onnx_dir, to_name))
                label = to_name
                valid_cnt += 1
            except Exception as e:
                print(f"Fail when seed={seed}")
                print(e)
                label = "FAILURE"

            time_diff = time.time() - tstart
            config_file.write(f"{time_diff:.5f},{label}\n")

            gen_cnt += 1
            config_file.flush()

            pbar.update(int(time.time() - start_time) - pbar.n)
            pbar.set_description(f"valid={valid_cnt},fail={gen_cnt-valid_cnt}")
            pbar.refresh()
        config_file.close()



import argparse
import random

from tqdm import tqdm

from nnsmith.graph_gen import random_model_gen, SymbolNet


def mknet(args):
    model_seed = random.getrandbits(32)
    gen, solution = random_model_gen(
        mode=args.mode, seed=model_seed, max_nodes=args.max_nodes, init_fp=True
    )
    net = SymbolNet(gen.abstract_graph, solution, alive_shapes=gen.alive_shapes)
    net.eval()
    return net, gen.num_op(), model_seed


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--max_nodes", type=int, default=20)
    parser.add_argument("--n_model", type=int, default=1000)
    parser.add_argument("--mode", type=str, default="random")
    args = parser.parse_args()

    n_invalid = 0
    with tqdm(range(args.n_model)) as pbar:
        for model_id in pbar:
            net, num_op, model_seed = mknet(args)
            net.check_intermediate_numeric = True
            _ = net(*net.get_random_inps(base="center", margin=1))
            n_invalid += net.invalid_found_last
            cur_model_size = model_id + 1
            pbar.set_description(
                f"invalid:{n_invalid}/{cur_model_size} = {100 * n_invalid/cur_model_size:.2f}%"
            )



from nnsmith.abstract.arith import nnsmith_lt
from nnsmith.abstract.extension import patch_requires


@patch_requires("global", "core.NCHWConv2d")
def limit_conv2d(self, _):
    # let the kernels to be > 3
    return [nnsmith_lt(3, self.kernel_h_size), nnsmith_lt(3, self.kernel_w_size)]



from nnsmith.filter import filter
from nnsmith.materialize import BugReport


@filter("test_fn")
def filter_fn(report: BugReport):
    return False  # Won't filter anything.


@filter("test_cls")
class FilterCls:
    def __call__(self, report: BugReport) -> bool:
        return False  # Won't filter anything.



from math import prod
from random import randint
from typing import List, Tuple, Union

from nnsmith.abstract.arith import *
from nnsmith.abstract.dtype import (
    DTYPE_GEN_ALL,
    DTYPE_GEN_COMPLEX,
    DTYPE_GEN_FLOATS,
    DTYPE_GEN_INTS,
    DTYPE_GEN_NON_BOOL,
    DType,
)
from nnsmith.abstract.op import (
    MatMul,
    ReduceBase,
    UnaryOpBase,
    mark_materialize,
    rank_from,
)
from nnsmith.abstract.tensor import AbsTensor
from nnsmith.error import ConstraintCheck


@mark_materialize("torch")
class Linear(UnaryOpBase):
    in_dtypes = [(DType.float32,)]
    out_dtypes = [(DType.float32,)]

    def __init__(self, ifeat: Union[int, z3.ExprRef], ofeat: Union[int, z3.ExprRef]):
        super().__init__()
        self.ifeat = ifeat
        self.ofeat = ofeat
        self.inp_ranks = [rank_from(1)]
        self.out_ranks = [rank_from(1)]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        assert len(input_shapes) == 1, "Linear only takes one input, but got {}".format(
            len(input_shapes)
        )
        return [
            AbsTensor(
                shape=[*input_shapes[0].shape[:-1], self.ofeat], dtype=DType.float32
            )
        ]

    def requires(self, input_shapes: List[AbsTensor]) -> List[z3.ExprRef]:
        ConstraintCheck.true(input_shapes[0].ndims >= 1)
        return [
            nnsmith_ge(self.ifeat, 1),
            nnsmith_ge(self.ofeat, 1),
            nnsmith_eq(input_shapes[0].shape[-1], self.ifeat),
        ]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(out_abs_tensor[0].ndims, DType.float32)]


@mark_materialize("torch")
class Flatten(UnaryOpBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_ALL]
    out_dtypes = [(i,) for i in DTYPE_GEN_ALL]

    def __init__(self):
        super().__init__()
        self.inp_ranks = [rank_from(1)]
        self.out_ranks = [(1,)]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        inp = input_shapes[0]
        return [
            AbsTensor(
                shape=[prod(inp.shape)],
                dtype=inp.dtype,
            )
        ]

    def requires(self, input_shapes):
        return []

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [(randint(0, 4), out_abs_tensor[0].dtype)]


@mark_materialize("torch")
class TorchReduceSum(ReduceBase):
    in_dtypes = [(i,) for i in DTYPE_GEN_NON_BOOL]
    out_dtypes = [(i,) for i in DTYPE_GEN_FLOATS + DTYPE_GEN_COMPLEX + [DType.int64]]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        output = super().type_transfer(input_shapes)
        if input_shapes[0].dtype in DTYPE_GEN_INTS:  # This is a PyTorch trick...
            output[0].dtype = DType.int64
        return output


@mark_materialize("torch")
class PTMatMul(MatMul):
    pass



from typing import List, Optional, Type

REQUIRES_PATCH = {}
ACTIVATED_PATCH = {}


class patch_requires:
    def __init__(self, tag: str, opname: str):
        self.tag = tag
        self.opname = opname

    def __call__(self, f):
        REQUIRES_PATCH.setdefault(self.tag, {}).setdefault(self.opname, []).append(f)
        return f


def activate_ext(
    opset: List[Type["AbsOpBase"]], factory: Optional["BackendFactory"] = None
):
    for op in opset:
        if "global" in REQUIRES_PATCH:
            ACTIVATED_PATCH.setdefault(op.name(), []).extend(
                REQUIRES_PATCH["global"].get(op.name(), [])
            )

        if factory is not None and factory.system_name in REQUIRES_PATCH:
            ACTIVATED_PATCH.setdefault(op.name(), []).extend(
                REQUIRES_PATCH[factory.system_name].get(op.name(), [])
            )



from os import PathLike
from typing import Dict, List

import numpy as np
import tensorflow as tf  # type: ignore
from multipledispatch import dispatch

from nnsmith.abstract.dtype import DType
from nnsmith.backends.factory import BackendCallable, BackendFactory
from nnsmith.materialize.tensorflow import TFModel, TFNetCallable


class TFLiteRunner:
    def __init__(self, tfnet_callable: TFNetCallable) -> None:
        self.tfnet_callable = tfnet_callable

    def __call__(self, input: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        return {k: np.array(v) for k, v in self.tfnet_callable(**input).items()}


class TFLiteFactory(BackendFactory):
    """Factory to build TFLite backend.
    Convertion Graph:
        TFModel & one concrete function
                    v  TFLiteConverter with configs
        TFLite model content in bytes -- (f.write) --> model.tflite file
                    v  tf.lite.Interpreter  <-- (f.read) <---
        TFLite Python Callable
    """

    def __init__(self, target, optmax, **kwargs) -> None:
        # https://github.com/tensorflow/tensorflow/issues/34536#issuecomment-565632906
        # TFLite doesn't support NVIDIA GPU.
        assert target != "cuda"
        super().__init__(target, optmax, **kwargs)

    @property
    def system_name(self) -> str:
        return "tflite"

    @property
    def version(self) -> str:
        return tf.__version__

    @dispatch(TFModel)
    def make_backend(
        self,
        model: TFModel,
    ) -> BackendCallable:
        """Create TFLite callable from a concrete function.
        TFModel is required because functions have a weak reference to Variables, which are stored in `tf.Module`.
        Ref: https://www.tensorflow.org/api_docs/python/tf/lite/TFLiteConverter#from_concrete_functions
        """
        return self.make_backend_from_content(self.make_content(model))

    def make_backend_from_path(
        self,
        path: PathLike,
    ) -> BackendCallable:
        """Create TFLite callable from path of a TF SavedModel.
        Ref: https://www.tensorflow.org/api_docs/python/tf/lite/TFLiteConverter#from_saved_model
        """
        return self.make_backend_from_content(
            self.make_content(path),
        )

    @dispatch(bytes)
    def make_backend_from_content(self, content: bytes) -> BackendCallable:
        # Ref: https://www.tensorflow.org/api_docs/python/tf/lite/Interpreter
        interpreter = tf.lite.Interpreter(model_content=content)
        return TFLiteRunner(interpreter.get_signature_runner())

    @dispatch(TFModel)
    def make_content(
        self,
        model: TFModel,
    ) -> bytes:
        """Create TFLite content from a concrete function.
        TFModel is required because functions have a weak reference to Variables, which are stored in `tf.Module`.
        Ref: https://www.tensorflow.org/api_docs/python/tf/lite/TFLiteConverter#from_concrete_functions
        """
        converter = tf.lite.TFLiteConverter.from_concrete_functions(
            funcs=[model.concrete_net()],
            trackable_obj=model.net,
        )
        return self._tflite_content_from_converter(converter)

    def make_content_with_func(
        self,
        model: TFModel,
        concrete_func: TFNetCallable,
    ) -> bytes:
        """Create TFLite content from a concrete function.
        TFModel is required because functions have a weak reference to Variables, which are stored in `tf.Module`.
        Ref: https://www.tensorflow.org/api_docs/python/tf/lite/TFLiteConverter#from_concrete_functions
        """
        converter = tf.lite.TFLiteConverter.from_concrete_functions(
            funcs=[concrete_func],
            trackable_obj=model.net,
        )
        return self._tflite_content_from_converter(converter)

    def make_backend_with_func(
        self,
        model: TFModel,
        concrete_func: TFNetCallable,
    ) -> BackendCallable:
        """Create TFLite callable from a concrete function.
        TFModel is required because functions have a weak reference to Variables, which are stored in `tf.Module`.
        Ref: https://www.tensorflow.org/api_docs/python/tf/lite/TFLiteConverter#from_concrete_functions
        """
        return self.make_backend_from_content(
            self.make_content_with_func(model, concrete_func)
        )

    @dispatch(str)
    def make_content(
        self,
        path: PathLike,
    ) -> bytes:
        """Create TFLite content from path of a saved model.
        Ref: https://www.tensorflow.org/api_docs/python/tf/lite/TFLiteConverter#from_saved_model
        """
        converter = tf.lite.TFLiteConverter.from_saved_model(path)
        return self._tflite_content_from_converter(converter)

    def _tflite_content_from_converter(
        self,
        converter: tf.lite.TFLiteConverter,
    ) -> bytes:
        """Configure TFLite converter and create callable from it.
        converter configuarations: https://www.tensorflow.org/api_docs/python/tf/lite/TFLiteConverter#attributes_1
        """
        # Ref: https://www.tensorflow.org/api_docs/python/tf/lite/TargetSpec
        converter.target_spec.supported_ops = [
            tf.lite.OpsSet.TFLITE_BUILTINS,  # enable TensorFlow Lite ops.
            tf.lite.OpsSet.SELECT_TF_OPS,  # enable TensorFlow ops.
        ]
        # Ref: https://www.tensorflow.org/api_docs/python/tf/lite/Optimize
        converter.optimizations = {
            tf.lite.Optimize.DEFAULT,
            tf.lite.Optimize.EXPERIMENTAL_SPARSITY,
        }
        # converter.allow_custom_ops = True
        tflite_bytes = converter.convert()
        return tflite_bytes

    def dump_backend(self, path: PathLike, content: bytes) -> None:
        with open(path, "wb") as f:
            f.write(content)

    def load_content(self, path: PathLike) -> bytes:
        with open(path, "rb") as f:
            return f.read()

    def load_backend(self, path: PathLike) -> BackendCallable:
        return self.make_backend_from_content(self.load_content(path))

    @classmethod
    def skip_dtypes(cls) -> List[DType]:
        # PyTorch-ONNX current does not support fully complex (e.g., scalar).
        # uint8 crash over Equal (revert after their fix).
        return [DType.uint8]



import logging
import os
import time
import warnings
from typing import Dict, Optional

import torch
from torch import nn

from nnsmith.abstract.dtype import DType
from nnsmith.abstract.op import AbsOpBase, Input
from nnsmith.abstract.tensor import AbsTensor
from nnsmith.error import ConstraintCheck, ConstraintError, SanityCheck
from nnsmith.gir import GraphIR
from nnsmith.logging import TORCH_LOG
from nnsmith.materialize.torch.forward import forward_fn
from nnsmith.materialize.torch.numeric import loss_fn, numeric_valid
from nnsmith.materialize.torch.proxy_grad import proxy_fn

__MB_LIM__ = 6 * 1024
__INPUT_FOUND_NAN_MSG__ = "[NaN] in model inputs!"
__INPUT_FOUND_INF_MSG__ = "[Inf] in model inputs!"
__ENABLE_RT_CHECK__ = os.getenv("NNSMITH_RT_CHECK", "0") == "1"


def check_type(op: AbsOpBase, tensors, is_input=True, msg=""):
    if __ENABLE_RT_CHECK__ and op is not None:
        like = op.input_like if is_input else op.output_like
        ioro = "input" if is_input else "output"

        assert len(like) == len(
            tensors
        ), f"{op}'s {ioro} has {len(like)} abs. input, but got {len(tensors)} real inputs."

        for i, ten in enumerate(tensors):
            ttype = AbsTensor(list(ten.shape), DType.from_torch(ten.dtype))
            assert (
                like[i] == ttype
            ), f"{msg} {ioro} abstract type != concrete {ttype} for {op} {op.input_like} -> {op.output_like}"


# Probablistically, sampling at positive domain is beneficial.
def random_tensor(shape, dtype: torch.dtype, margin=4, base=5, use_cuda=False):
    # center: -margin ~ 0 ~ +margin
    dev = torch.device("cuda" if use_cuda else "cpu")
    if base == "center":
        base = -margin / 2
    else:
        assert isinstance(base, int) or isinstance(base, float)

    fp_tensor = base + (torch.rand(shape, device=dev) - 0.5) * margin

    if dtype.is_floating_point:
        return fp_tensor.to(dtype)
    else:
        return torch.round(fp_tensor).to(dtype)


class SymbolNet(nn.Module):
    def __init__(
        self,
        ir: GraphIR,
        record_intermediate=False,
        use_gradient=False,
        print_grad=0,
    ):
        super(SymbolNet, self).__init__()
        self.print_grad = print_grad
        # <TorchFunc, <keys -> inputs>, <keys -> outputs>, original op>
        self.instructions = []
        self.n_vulnerable_op = 0

        self.proxy_enabled_ = False

        # keep track of layers and weights so that the tracing can work properly
        self.mlist = nn.ModuleList()
        # whether or not to register intermediate tensors as output tensors. Useful (at least) for checking nan
        self.record_intermediate = record_intermediate
        self._device = None

        self.ir = ir

        for inst in self.ir.insts:
            if not isinstance(inst.iexpr.op, Input):
                torch_fn = forward_fn(inst.iexpr.op)
                SanityCheck.true(torch_fn is not None, f"Bad impl for {inst.iexpr.op}")
                if isinstance(torch_fn, nn.Module):
                    self.mlist.append(torch_fn)
                self.instructions.append(
                    (torch_fn, inst.iexpr.args, inst.retvals(), inst.iexpr.op)
                )

                if loss_fn.dispatch(type(inst.iexpr.op)):
                    self.n_vulnerable_op += 1

        # the order follows `input_keys`

        self.input_map = {iname: self.ir.vars[iname] for iname in self.ir.input_var()}
        self.output_map = {oname: self.ir.vars[oname] for oname in self.ir.leaf_var()}

        self.first_run = True

        self.use_gradient = use_gradient
        if use_gradient:
            self.enable_training()
        self.check_intermediate_numeric = False
        self.invalid_found_last = None

    def to(self, device):
        self._device = device
        return super().to(device)

    @property
    def device(self):
        return self._device

    @property
    def input_like(self):
        return self.input_map

    @property
    def output_like(self):
        return self.output_map

    @property
    def proxy_enabled(self):
        return self.proxy_enabled_

    def enable_proxy_grad(self):
        for i, inst in enumerate(self.instructions):
            _, inputs, outputs, op = inst
            if proxy_fn.dispatch(type(op)):  # has proxy
                self.instructions[i] = (proxy_fn(op), inputs, outputs, op)

        self.proxy_enabled_ = True

    def disable_proxy_grad(self):
        for i, inst in enumerate(self.instructions):
            _, inputs, outputs, op = inst
            if proxy_fn.dispatch(type(op)):  # has proxy
                self.instructions[i] = (forward_fn(op), inputs, outputs, op)

        self.proxy_enabled_ = False

    def get_params(self):
        return sum([i["params"] for i in self.optimizer.param_groups], [])

    def _zero_grad(self):
        for p in self.get_params():
            p.grad = None

    def backward(self):
        if self.loss is not None:
            self._zero_grad()
            params = self.get_params()
            loss_name, l = self.loss
            l.backward()

            with torch.no_grad():
                for param in self.parameters():
                    param.copy_(
                        torch.where(
                            param.isnan().logical_or(param.isinf()),
                            random_tensor(
                                shape=param.shape, dtype=param.dtype.torch()
                            ).to(param.device),
                            param,
                        )
                    )

            if self.print_grad >= 2:
                for name, i in self.interm_grad:
                    msg = (
                        f"{i.grad.min()} ~ {i.grad.max()} ~ {i.grad.mean()}"
                        if i.grad is not None
                        else "None"
                    )
                    TORCH_LOG.info(
                        f"Iter {self.iter_num} [{loss_name}] {name} grad: {msg}"
                    )

            nonzero = False
            with torch.no_grad():
                for i, p in enumerate(params):
                    if p.grad is not None and torch.any(p.grad != 0):
                        nonzero = True
                        break  # As long as there's non-zero grad.
            ConstraintCheck.true(
                nonzero, "Gradients are all zero. Cannot make progress."
            )

            torch.nn.utils.clip_grad_norm_(self.to_train, 1e-1)
            self.optimizer.step()

    def training_reset(self):
        self.loss = None
        self.stop_updating_loss = False

    def stop_training(self):
        self.use_gradient = False
        self.loss = None

    def enable_training(self, extra_trainable: Dict[str, torch.Tensor] = {}):
        self.use_gradient = True
        self.to_train = []
        for t in extra_trainable.values():
            self.to_train.append(t)
        for t in self.parameters():
            self.to_train.append(t)
        self.optimizer = torch.optim.Adam(self.to_train, lr=5e-1)
        self.training_reset()

    def reset_optimizer(self):
        self.optimizer = torch.optim.Adam(self.to_train, lr=5e-1)

    def get_random_inps(self, **kwargs) -> Dict[str, torch.Tensor]:
        # center: -margin ~ 0 ~ +margin
        inputs = {}
        for key, abs_tensor in self.input_like.items():
            inputs[key] = random_tensor(
                abs_tensor.shape, abs_tensor.dtype.torch(), **kwargs
            )

        return inputs

    def grad_input_gen(
        self,
        init_tensors: Dict[str, torch.Tensor],
        use_cuda=False,
        max_time=None,
        **kwargs,
    ) -> Optional[Dict[str, torch.Tensor]]:
        # TODO: trim the param. max_iter is not used; remove getenv
        if max_time is None:
            max_time = float(os.getenv("NNSMITH_GRAD_TIME", 0.5))

        inputs: Dict[str, torch.Tensor] = {}
        for key, tensor in init_tensors.items():
            if tensor.data.dtype.is_floating_point:
                inputs[key] = torch.nn.parameter.Parameter(tensor.data.clone())
            else:
                inputs[key] = tensor.data

        self.enable_training(extra_trainable=inputs)

        last_check_intermediate_numeric = self.check_intermediate_numeric
        self.check_intermediate_numeric = True

        if use_cuda:
            for k in inputs:
                inputs[k] = inputs[k].cuda()
            self.use_cuda()

        sat_inputs = None
        st = time.time()
        self.iter_num = 0
        self.cur_loss_name = None
        while time.time() - st < max_time:
            self.training_reset()
            self.iter_num += 1

            try:
                _ = self(**inputs)
                if self.invalid_found_last:  # need_to_train
                    self.backward()
                else:
                    sat_inputs = {k: v.data for k, v in inputs.items()}
                    break
            except ConstraintError as e:
                if __INPUT_FOUND_INF_MSG__ in str(e) or __INPUT_FOUND_NAN_MSG__ in str(
                    e
                ):
                    # flush NaN/Inf in inputs
                    with torch.no_grad():
                        for inp in inputs.values():
                            inp.copy_(
                                torch.where(
                                    inp.isnan().logical_or(inp.isinf()),
                                    random_tensor(
                                        shape=inp.shape, dtype=inp.dtype.torch()
                                    ).to(inp.device),
                                    inp,
                                )
                            )
                    continue
                TORCH_LOG.debug(e)
                break

        self.stop_training()
        if sat_inputs is None:
            TORCH_LOG.debug("[grad] no valid range found!!!")

        self.check_intermediate_numeric = last_check_intermediate_numeric
        return sat_inputs

    def use_cuda(self):
        self.cuda()

    @torch.jit.ignore
    def debug_numeric(self, tensor_map):
        with warnings.catch_warnings():  # just shutup.
            warnings.simplefilter("ignore")
            ConstraintCheck.true(
                not any([torch.isinf(op).any() for _, op in tensor_map.items()]),
                __INPUT_FOUND_INF_MSG__,
            )
            ConstraintCheck.true(
                not any([torch.isnan(op).any() for _, op in tensor_map.items()]),
                __INPUT_FOUND_NAN_MSG__,
            )

    def forward(self, *args, **kwargs):
        self.differentiable = True

        tensor_map: Dict[str, torch.Tensor] = {}

        if len(args) == len(self.input_map):
            for i, key in enumerate(self.ir.input_var()):
                tensor_map[key] = args[i]
        elif len(kwargs) == len(self.input_map):
            for ir_key in self.input_map:
                tensor_map[ir_key] = kwargs[ir_key]
        else:
            raise ValueError("Either user args only or kwargs only")

        self.debug_numeric(tensor_map)

        self.invalid_found_last = False

        self.interm_grad = []

        # LOG.
        if self.print_grad >= 2:
            for k, v in tensor_map.items():
                if v.requires_grad:
                    self.interm_grad.append((k, v))

            for i, p in enumerate(self.parameters()):
                if p.requires_grad:
                    self.interm_grad.append((f"p_{i}", p))

        for stmt_idx, (inst, inps, outs, op) in enumerate(self.instructions):
            input_tensors = [tensor_map[idx] for idx in inps]

            check_type(op, input_tensors, is_input=True, msg="input")

            # REAL FORWARD.
            output_tensors = inst(*input_tensors)
            if not isinstance(output_tensors, list):
                output_tensors = [output_tensors]

            check_type(op, output_tensors, is_input=False, msg="output")

            for i, out_key in enumerate(outs):
                # put values back to tensor_map.
                tensor_map[out_key] = output_tensors[i]
                # Check differentiability.
                self.differentiable &= output_tensors[i].grad_fn is not None
                # TODO(@ganler): optimize: unref tensors that are not going to be used anymore.

            # LOG.
            if TORCH_LOG.isEnabledFor(logging.DEBUG):
                TORCH_LOG.debug(f">> statment {stmt_idx}")
                for inp_i, i in enumerate(input_tensors):
                    TORCH_LOG.debug(f"  (shape={i.shape} dtype={i.dtype})")
                    TORCH_LOG.debug(
                        f"[inp]@{inp_i} :: {i.min().data:.5f} ~ {i.max().data:.5f}"
                    )
                for out_i, o in enumerate(output_tensors):
                    TORCH_LOG.debug(f"  (shape={o.shape} dtype={o.dtype})")
                    TORCH_LOG.debug(
                        f"[out]@{out_i} :: {o.min().data:.5f} ~ {o.max().data:.5f}"
                    )

            if self.print_grad >= 2:
                if output_tensors[0].requires_grad:
                    for i in range(len(output_tensors)):
                        output_tensors[i].retain_grad()
                        self.interm_grad.append((f"{op}{i}", output_tensors[i]))

            if self.check_intermediate_numeric or (
                self.use_gradient and not self.stop_updating_loss
            ):
                if loss_fn.dispatch(type(op)) is not None:
                    loss = loss_fn(op)(*input_tensors)
                    if not isinstance(loss, tuple):
                        loss = ("", loss)  # loss suffix, loss
                    vul_op_loss = loss
                else:
                    vul_op_loss = None
                self.invalid_found_last |= not numeric_valid(output_tensors)

                if self.invalid_found_last and (
                    self.use_gradient and not self.stop_updating_loss
                ):
                    if self.print_grad >= 1:
                        for inp_i, inp in enumerate(input_tensors):
                            TORCH_LOG.info(
                                f"[inp]@{inp_i} :: {inp.min().data:.5f} ~ {inp.max().data:.5f}"
                            )

                    ConstraintCheck.true(
                        vul_op_loss is not None,
                        f"op={op} has no `torch_loss` but produces NaN or INF!",
                    )
                    # TODO: some less vulnerable ops (like Mul) may also trigger Inf and will crash the process.
                    # Given its low chance of happening, ignore it for now.
                    loss_suf, l = vul_op_loss
                    msg = f"loss_{loss_suf}: {l.min().data:.3f} ~ {l.max().data:.3f} ~ {torch.sum((l > 0) * l).item()}"
                    if self.print_grad >= 1:
                        TORCH_LOG.info(
                            f"Iter #{self.iter_num} [NaN/Inf] in outputs ~ {op} :: {msg}"
                        )

                    ConstraintCheck.true(
                        torch.all(l > 0),
                        f"`{op}` outputs NaN/INF found non-positive loss!",
                    )
                    loss_name = f"{op}_{loss_suf}"
                    ConstraintCheck.true(self.loss is None, "Multiple loss detected!")
                    self.loss = loss_name, torch.sum((l > 0) * l)
                    if loss_name != self.cur_loss_name:
                        self.reset_optimizer()
                        self.cur_loss_name = loss_name

                    self.stop_updating_loss = True
                    return output_tensors

        self.first_run = False
        return tuple(tensor_map[key] for key in self.output_map)



import pickle
from abc import ABC, abstractmethod
from os import PathLike
from typing import Dict, List, Type

import torch

from nnsmith.abstract.op import AbsOpBase, AbsTensor
from nnsmith.gir import GraphIR
from nnsmith.materialize import Model, Oracle
from nnsmith.materialize.torch.forward import ALL_TORCH_OPS
from nnsmith.materialize.torch.input_gen import PracticalHybridSearch
from nnsmith.materialize.torch.symbolnet import SymbolNet
from nnsmith.util import register_seed_setter


class TorchModel(Model, ABC):
    def __init__(self) -> None:
        super().__init__()
        self.torch_model: SymbolNet = None
        self.sat_inputs = None

    @classmethod
    @abstractmethod
    def device(cls) -> torch.device:
        pass

    @property
    def version(self) -> str:
        return torch.__version__

    @classmethod
    def from_gir(cls: Type["TorchModel"], ir: GraphIR, **kwargs) -> "TorchModel":
        ret = cls()
        ret.torch_model = SymbolNet(ir, **kwargs).to(cls.device())
        return ret

    @staticmethod
    def gir_name() -> str:
        return "gir.pkl"

    def refine_weights(self) -> None:
        self.torch_model.enable_proxy_grad()
        use_cuda = self.device().type == "cuda"
        searcher = PracticalHybridSearch(self.torch_model, use_cuda=use_cuda)
        # TODO(@ganler): Can we directly get both inputs and outputs?
        _, inputs = searcher.search(
            max_time_ms=20,
            max_sample=2,
        )
        if inputs:
            self.sat_inputs = inputs
        self.torch_model.disable_proxy_grad()

    def make_oracle(self) -> Oracle:
        with torch.no_grad():
            self.torch_model.eval()
            # fall back to random inputs if no solution is found.
            if self.sat_inputs is None:
                inputs = self.torch_model.get_random_inps()
            else:
                inputs = self.sat_inputs
            inputs = {k: v.to(self.device()) for k, v in inputs.items()}
            outputs = self.torch_model.forward(**inputs)

        # numpyify
        input_dict = {k: v.cpu().detach().numpy() for k, v in inputs.items()}
        output_dict = {}
        for oname, val in zip(self.output_like.keys(), outputs):
            output_dict[oname] = val.cpu().detach().numpy()

        return Oracle(input_dict, output_dict, provider="torch[cpu] eager")

    def dump(self, path: PathLike):
        torch.save(self.torch_model.state_dict(), path)
        gir_path = path.replace(
            TorchModel.name_prefix() + TorchModel.name_suffix(),
            TorchModel.gir_name(),
        )
        with open(gir_path, "wb") as f:
            pickle.dump(self.torch_model.ir, f)

    @classmethod
    def load(cls, path: PathLike) -> "TorchModel":
        ret = cls()
        gir_path = path.replace(
            cls.name_prefix() + cls.name_suffix(),
            cls.gir_name(),
        )
        with open(gir_path, "rb") as f:
            ir = pickle.load(f)
        torch_model = SymbolNet(ir).to(cls.device())
        torch_model.load_state_dict(torch.load(path), strict=False)
        ret.torch_model = torch_model
        return ret

    @staticmethod
    def name_suffix() -> str:
        return ".pth"

    @property
    def input_like(self) -> Dict[str, AbsTensor]:
        return self.torch_model.input_like

    @property
    def output_like(self) -> Dict[str, AbsTensor]:
        return self.torch_model.output_like

    @property
    def native_model(self) -> SymbolNet:
        return self.torch_model

    @staticmethod
    def operators() -> List[Type[AbsOpBase]]:
        return ALL_TORCH_OPS

    @staticmethod
    def add_seed_setter() -> None:
        register_seed_setter("torch", torch.manual_seed, overwrite=True)


class TorchModelCPU(TorchModel):
    @classmethod
    def device(cls) -> torch.device:
        return torch.device("cpu")


class TorchModelCUDA(TorchModel):
    @classmethod
    def device(cls) -> torch.device:
        return torch.device("cuda")



import pytest
import torch

from nnsmith.abstract.dtype import DType
from nnsmith.backends import BackendFactory
from nnsmith.graph_gen import model_gen
from nnsmith.materialize import Model, TestCase
from nnsmith.narrow_spec import auto_opconfig, auto_opset

TestCase.__test__ = False  # supress PyTest warning


def test_narrow_spec_cache_make_and_reload():
    factory = BackendFactory.init("torchjit", target="cpu", optmax=True)
    ModelType = Model.init("torch")
    opset_lhs = auto_opconfig(ModelType, factory)
    assert opset_lhs, "Should not be empty... Something must go wrong."
    opset_rhs = auto_opconfig(ModelType, factory)
    assert opset_lhs == opset_rhs

    # Assert types
    assert isinstance(opset_lhs["core.ReLU"].in_dtypes[0][0], DType)

    # Assert Dictionary Type Equality
    assert type(opset_lhs) == type(opset_rhs)
    assert type(opset_lhs["core.ReLU"]) == type(opset_rhs["core.ReLU"])
    assert type(opset_lhs["core.ReLU"].in_dtypes[0][0]) == type(
        opset_rhs["core.ReLU"].in_dtypes[0][0]
    )


def test_synthesized_tf_model(tmp_path):
    d = tmp_path / "test_torchjit"
    d.mkdir()

    targets = ["cpu"]
    if torch.cuda.is_available():
        targets.append("cuda")

    for target in targets:
        factory = BackendFactory.init("torchjit", target=target, optmax=False)

        ModelType = Model.init("torch", backend_target=target)

        gen = model_gen(
            opset=auto_opset(ModelType, factory),
            seed=23132,
            max_nodes=1,
        )  # One op should not be easily wrong... I guess.

        model = ModelType.from_gir(gen.make_concrete())

        oracle = model.make_oracle()

        testcase = TestCase(model, oracle)
        assert factory.verify_testcase(testcase) is None
        testcase.dump(root_folder=d)



import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib_venn import _venn2, _venn3, venn2, venn3

SMALL_SIZE = 10
MEDIUM_SIZE = 14
BIGGER_SIZE = 18

plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
plt.rc("axes", titlesize=MEDIUM_SIZE)  # fontsize of the axes title
plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
plt.rc("ytick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
plt.rc("legend", fontsize=MEDIUM_SIZE - 1)  # legend fontsize
plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title

MIN_FAC_TWO = None
MIN_FAC = None

plt.rc("text", usetex=True)
plt.rc("text.latex", preamble=r"\usepackage{xfrac}")


class Ploter:
    def __init__(self, cov_lim=None, use_pdf=False, one_plot=False, scale=1) -> None:
        # cov / time, cov / iteration, iteration / time
        fig0, axs0 = plt.subplots(1, 1, constrained_layout=True, figsize=(6, 3.5))
        fig1, axs1 = plt.subplots(1, 1, constrained_layout=True, figsize=(5, 3))

        self.one_plot = one_plot
        self.fig = [fig0, fig1]
        self.axs = [axs0, axs1]
        self.cov_lim = cov_lim
        self.cov_maxes = []
        self.xspan = 0
        self.use_pdf = use_pdf
        self.scale = scale

    def add(self, data, name=None):
        df = np.array(data)
        df[:, 2] = df[:, 2] / self.scale

        LW = 2
        MARKER_SIZE = 10
        N_MARKER = 8
        MARKERS = ["d", "^", "p", "*"]
        LS = ":"
        COLORS = ["dodgerblue", "violet", "coral"]

        # make it even over time
        markevery = np.zeros_like(df[:, 0], dtype=bool)
        step = int(df[:, 0].max() / N_MARKER)
        offset = step
        for _ in range(N_MARKER):
            idx = list(map(lambda i: i >= offset, df[:, 0])).index(True)
            markevery[idx] = True
            offset += step

        # markevery = int(len(df) / N_MARKER)
        marker = MARKERS[len(self.cov_maxes) % len(MARKERS)]
        color = COLORS[len(self.cov_maxes) % len(MARKERS)]

        # linestyle=LS, marker=marker, markevery=markevery, markersize=MARKER_SIZE, alpha=ALPHA, lw=LW
        style_kw = {
            "linestyle": LS,
            "marker": marker,
            "markevery": markevery,
            "markersize": MARKER_SIZE,
            "lw": LW,
            "color": color,
            "markeredgecolor": "k",
            "markeredgewidth": 1.5,
        }

        self.axs[0].plot(df[:, 0], df[:, 2], label=name, **style_kw)  # cov / time
        print(
            f"----> max cov {df[:, 2].max() * self.scale} + max tests {int(df[:, 1].max())}"
        )

        self.axs[1].plot(df[:, 1], df[:, 2], label=name, **style_kw)  # cov / iteration

        self.xspan = max(self.xspan, df[-1, 0])

        self.cov_maxes.append(df[:, 2].max())

    def plot(self, save="cov", cov_lim=None, loc=0):
        for ax in self.axs:
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles[::-1], labels[::-1], loc=loc)

        self.cov_maxes = sorted(self.cov_maxes)

        if len(self.cov_maxes) > 1:
            for rk in range(1, len(self.cov_maxes)):
                best = self.cov_maxes[-1]
                cur = self.cov_maxes[-(rk + 1)]
                print(
                    f"==> Best one is [{best} / {cur}] = **{best / cur:.2f}x** better than the NO. {rk + 1} baseline."
                )

        cov_max = max(self.cov_maxes)
        cov_min = min(self.cov_maxes)

        top_lim = cov_max + (cov_max - cov_min) * 0.5

        if cov_lim is not None:
            ann_size = BIGGER_SIZE + 4
            max_cov_unscale = int(cov_max * self.scale)
            self.axs[0].annotate(
                f"$\\sfrac{{{max_cov_unscale}_\\mathbf{{best}}}}{{{int(cov_lim)}_\\mathbf{{max}}}}$ = \\textbf{{{max_cov_unscale / cov_lim * 100 :.1f}\%}}",
                xy=(0.7, 0.1),
                xycoords="axes fraction",
                ha="right",
                fontsize=ann_size,
                bbox=dict(boxstyle="round", fc="w", alpha=0.6),
            )

        if self.cov_lim is not None:
            self.axs[0].set_ylim(bottom=self.cov_lim, top=top_lim)
            self.axs[1].set_ylim(bottom=self.cov_lim, top=top_lim)
        else:
            self.axs[0].set_ylim(bottom=cov_min * MIN_FAC, top=top_lim)
            self.axs[1].set_ylim(bottom=cov_min * MIN_FAC, top=top_lim)

        ylabel = "\# Coverage"
        if self.scale != 1:
            ylabel = f"\# Coverage ({self.scale} branches)"
        self.axs[0].set(xlabel="Time (Second)", ylabel=ylabel)
        self.axs[0].grid(alpha=0.5, ls=":")

        self.axs[1].set(xlabel="\# Iteration", ylabel=ylabel)
        self.axs[1].grid(alpha=0.5, ls=":")

        if self.use_pdf:
            self.fig[0].savefig(save + "-time.pdf")
            self.fig[1].savefig(save + "-iter.pdf")
        self.fig[0].savefig(save + "-time.png")
        self.fig[1].savefig(save + "-iter.png")


def cov_summerize(data, pass_filter=None, tlimit=None, branch_only=True, gen_time=None):
    model_total = 0

    branch_by_time = [[0, 0, 0]]
    final_bf = 0

    line_by_time = [[0, 0, 0]]
    final_lf = 0

    for time, value in data.items():
        lf = 0
        bf = 0

        n_model = value["n_model"]
        cov = value["merged_cov"]
        model_total += n_model
        if gen_time is not None:
            time -= gen_time[0][:model_total].sum()
        line_cov = 0
        branch_cov = 0
        for fname in cov:
            if pass_filter is not None and not pass_filter(fname):
                continue

            branch_cov += len(cov[fname]["branches"])
            bf += cov[fname]["bf"]

            if not branch_only:
                line_cov += len(cov[fname]["lines"])
                lf += cov[fname]["lf"]

        if not branch_only:
            line_by_time.append([time, model_total, line_cov])

        branch_by_time.append([time, model_total, branch_cov])

        final_lf = max(final_lf, lf)
        final_bf = max(final_bf, bf)

        if tlimit is not None and time > tlimit:
            break
    return line_by_time, branch_by_time, (final_lf, final_bf)


def tvm_pass_filter(fname):
    if "relay/transforms" in fname:
        return True
    elif "src/tir/transforms" in fname:
        return True
    elif "src/ir/transform.cc" in fname:
        return True

    return False


def ort_pass_filter(fname):
    return "onnxruntime/core/optimizer/" in fname


def tvm_arith_filter(fname):
    return "arith" in fname


def plot_one_round(
    folder,
    data,
    pass_filter=None,
    fuzz_tags=None,
    target_tag="",
    tlimit=None,
    pdf=False,
    pass_tag="",
    venn=False,
    scale=1,
):
    branch_ploter = Ploter(use_pdf=pdf, scale=scale)

    assert fuzz_tags is not None
    if pass_filter is not None:
        assert pass_tag != ""

    # Due to lcov, diff lcov's total cov might be slightly different.
    # We took the max.
    lf = 0
    bf = 0

    for idx, (k, v) in enumerate(data.items()):
        _, branch_by_time, (lf_, bf_) = cov_summerize(
            v, tlimit=tlimit, pass_filter=pass_filter
        )
        branch_ploter.add(data=branch_by_time, name=fuzz_tags[idx])

        lf = max(lf, lf_)
        bf = max(bf, bf_)

    branch_ploter.plot(
        save=os.path.join(folder, target_tag + pass_tag + "branch_cov"),
        cov_lim=bf,
        loc=7 if ("ort" in target_tag) else 0,
    )

    if not venn:
        return

    # venn graph plot
    branch_cov_sets = []
    for _, v in data.items():
        last_key = sorted(list(v.keys()))[-1]
        # file -> {lines, branches}
        final_cov = v[last_key]["merged_cov"]
        branch_set = set()
        for fname in final_cov:
            if pass_filter is not None and not pass_filter(fname):
                continue
            brset = set([fname + br for br in final_cov[fname]["branches"]])
            branch_set.update(brset)
        branch_cov_sets.append(branch_set)

    if len(branch_cov_sets) != 1:
        plt.clf()
        if len(branch_cov_sets) == 2:
            plt.figure(figsize=(3.5, 2.5), constrained_layout=True)
            ks = ["10", "01", "11"]
            sets = {}
            total_covs = [len(s) for s in branch_cov_sets]
            for k, val in zip(ks, _venn2.compute_venn2_subsets(*branch_cov_sets)):
                sets[k] = val
            v = venn2(
                subsets=(5, 5, 2),
                set_labels=[f"{{{t}}}\n({c})" for t, c in zip(fuzz_tags, total_covs)],
            )

            lb = v.get_label_by_id("A")
            x, y = lb.get_position()
            lb.set_position((x + 0.07, y + 0.4))

            lb = v.get_label_by_id("B")
            x, y = lb.get_position()
            lb.set_position((x - 0.07, y + 0.4))

            for id in ["11"]:
                if v.get_label_by_id(id):
                    v.get_label_by_id(id).set_text(sets[id])
                    v.get_label_by_id(id).set_fontsize(MEDIUM_SIZE + 2)
                    v.get_patch_by_id(id).set_alpha(0.15)
                    v.get_patch_by_id(id).set_facecolor("royalblue")

            hatches = ["|", "\\"]
            # circles = ['dodgerblue', 'MediumVioletRed', 'coral', 'white'] # colorful.
            # fcolors = ['violet', 'navajowhite'] # for binning
            fcolors = ["lightgreen", "navajowhite"]
            for idx, id in enumerate(["10", "01"]):
                if sets[id] == 0:
                    continue
                cnt = sets[id]
                v.get_label_by_id(id).set_text(f"\\textbf{{{cnt}}}")
                v.get_label_by_id(id).set_fontsize(BIGGER_SIZE + 4)
                v.get_patch_by_id(id).set_edgecolor("gray")
                v.get_patch_by_id(id).set_hatch(hatches[idx])
                v.get_patch_by_id(id).set_facecolor(fcolors[idx])

        elif len(branch_cov_sets) == 3:
            plt.figure(figsize=(4.5, 3.5), constrained_layout=True)
            ks = ["100", "010", "110", "001", "101", "011", "111"]
            sets = {}
            total_covs = [len(s) for s in branch_cov_sets]
            for k, val in zip(ks, _venn3.compute_venn3_subsets(*branch_cov_sets)):
                sets[k] = val
            v = venn3(
                subsets=(6, 6, 3, 6, 3, 3, 4.5),
                set_labels=[f"{t}\n({c})" for t, c in zip(fuzz_tags, total_covs)],
            )

            lb = v.get_label_by_id("A")
            x, y = lb.get_position()
            lb.set_position((x - 0.2, y - 0.3))

            lb = v.get_label_by_id("B")
            x, y = lb.get_position()
            lb.set_position((x + 0.2, y - 0.3))

            lb = v.get_label_by_id("C")
            x, y = lb.get_position()
            lb.set_position((x + 0.6, y + 0.3))

            for id in ["110", "101", "011"]:
                if v.get_label_by_id(id):
                    v.get_label_by_id(id).set_text(sets[id])
                    v.get_label_by_id(id).set_fontsize(MEDIUM_SIZE)
                    v.get_patch_by_id(id).set_alpha(0.15)

            hatches = ["*", ".", "\\"]
            # circles = ['dodgerblue', 'MediumVioletRed', 'coral', 'white'] # colorful.
            circles = ["k"] * 4  # chill man!
            fcolors = ["lightblue", "violet", "navajowhite"]
            for idx, id in enumerate(["100", "010", "001", "111"]):
                if sets[id] == 0:
                    continue
                cnt = sets[id]
                v.get_label_by_id(id).set_text(f"\\textbf{{{cnt}}}")
                v.get_label_by_id(id).set_fontsize(BIGGER_SIZE + 4)
                v.get_patch_by_id(id).set_edgecolor(circles[idx])
                if id != "111":
                    v.get_patch_by_id(id).set_hatch(hatches[idx])
                    v.get_patch_by_id(id).set_facecolor(fcolors[idx])
                    # v.get_patch_by_id(id).set_linewidth(2)
                else:
                    v.get_patch_by_id(id).set_alpha(0.2)

        for text in v.set_labels:
            text.set_fontsize(BIGGER_SIZE - 1)

    plt.savefig(
        f'{os.path.join(folder, target_tag + pass_tag + "br_cov_venn")}.png',
        bbox_inches="tight",
    )
    if pdf:
        plt.savefig(
            f'{os.path.join(folder, target_tag + pass_tag + "br_cov_venn")}.pdf',
            bbox_inches="tight",
        )
    plt.close()


if "__main__" == __name__:
    import argparse
    import pickle

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f", "--folders", type=str, nargs="+", help="bug report folder"
    )
    parser.add_argument("--tags", type=str, nargs="+", help="tags")
    parser.add_argument(
        "-o", "--output", type=str, default="results", help="results folder"
    )
    parser.add_argument("-t", "--tlimit", type=int, default=4 * 3600, help="time limit")
    parser.add_argument("--tvm", action="store_true", help="use tvm")
    parser.add_argument("--ort", action="store_true", help="use ort")
    parser.add_argument("--pdf", action="store_true", help="use pdf as well")
    parser.add_argument("--venn", action="store_true", help="plot venn")
    args = parser.parse_args()

    if args.tags is None:
        args.tags = [os.path.split(f)[-1].split("-")[0] for f in args.folders]
        # args.tags = [os.path.split(os.path.split(f)[-2])[-1]
        #              for f in args.folders]
    else:
        assert len(args.tags) == len(args.folders)

    if not os.path.exists(args.output):
        os.mkdir(args.output)

    pass_filter = None
    target_tag = ""
    if args.tvm:
        pass_filter = tvm_pass_filter
        target_tag = "tvm_"
    elif args.ort:
        pass_filter = ort_pass_filter
        target_tag = "ort_"
    else:
        print(f"[WARNING] No pass filter is used (use --tvm or --ort)")

    arith_filter = None
    if args.tvm:
        arith_filter = tvm_arith_filter
        MIN_FAC = 0.87
    elif args.ort:
        arith_filter = None
        MIN_FAC = 0.7
    else:
        print(f"[WARNING] No pass filter is used (use --tvm or --ort)")

    data = {}
    gen_time = None
    for f in args.folders:
        with open(os.path.join(f, "merged_cov.pkl"), "rb") as fp:
            data[f] = pickle.load(fp)

    if pass_filter is not None:
        plot_one_round(
            folder=args.output,
            data=data,
            pass_filter=pass_filter,
            pass_tag="opt_",
            scale=100,
            tlimit=args.tlimit,
            fuzz_tags=args.tags,
            target_tag=target_tag,
            pdf=args.pdf,
            venn=args.venn,
        )
    plot_one_round(
        folder=args.output,
        data=data,
        pass_filter=None,
        tlimit=args.tlimit,
        scale=1000,
        fuzz_tags=args.tags,
        target_tag=target_tag,
        pdf=args.pdf,
        venn=args.venn,
    )  # no pass



import numpy as np
import pytest
import torch

from nnsmith.graph_gen import model_gen
from nnsmith.materialize import Model, Oracle, TestCase
from nnsmith.narrow_spec import auto_opset

TestCase.__test__ = False  # supress PyTest warning


def compare_two_oracle(src: Oracle, loaded: Oracle):
    assert len(loaded.input) == len(src.input)
    assert len(loaded.output) == len(src.output)
    for k, v in loaded.input.items():
        assert np.allclose(v, src.input[k], equal_nan=True)
    for k, v in loaded.output.items():
        assert np.allclose(v, src.output[k], equal_nan=True)


def test_onnx_load_dump(tmp_path):
    d = tmp_path / "test_onnx_load_dump"
    d.mkdir()

    ONNXModelCPU = Model.init("onnx")

    gen = model_gen(
        opset=auto_opset(ONNXModelCPU),
        seed=54341,
        max_nodes=5,
    )

    model = ONNXModelCPU.from_gir(gen.make_concrete())

    assert model.with_torch

    model.refine_weights()  # either random generated or gradient-based.
    oracle = model.make_oracle()

    testcase = TestCase(model, oracle)
    testcase.dump(root_folder=d)

    loaded_testcase = TestCase.load(model_type=type(model), root_folder=d)

    # check oracle
    compare_two_oracle(oracle, loaded_testcase.oracle)

    loaded_model = loaded_testcase.model.torch_model
    loaded_model.sat_inputs = {k: torch.from_numpy(v) for k, v in oracle.input.items()}
    rerun_oracle = loaded_model.make_oracle()
    compare_two_oracle(oracle, rerun_oracle)


def test_bug_report_load_dump(tmp_path):
    d = tmp_path / "test_onnx_load_dump"
    d.mkdir()

    ONNXModelCPU = Model.init("onnx")
    gen = model_gen(
        opset=auto_opset(ONNXModelCPU),
        seed=5341,
        max_nodes=5,
    )

    model = ONNXModelCPU.from_gir(gen.make_concrete())

    assert model.with_torch

    model.refine_weights()  # either random generated or gradient-based.
    oracle = model.make_oracle()

    testcase = TestCase(model, oracle)
    testcase.dump(root_folder=d)

    loaded_testcase = TestCase.load(model_type=type(model), root_folder=d)

    # check oracle
    compare_two_oracle(oracle, loaded_testcase.oracle)

    loaded_model = loaded_testcase.model.torch_model
    loaded_model.sat_inputs = {k: torch.from_numpy(v) for k, v in oracle.input.items()}
    rerun_oracle = loaded_model.make_oracle()
    compare_two_oracle(oracle, rerun_oracle)



from nnsmith.graph_gen import random_model_gen, SymbolNet
from nnsmith.materialize.torch.input_gen import PracticalHybridSearch
from nnsmith.materialize.onnx.export import torch2onnx
from nnsmith.dtype_test import rewrite_op_dtype
from nnsmith.abstract.op import ALL_OP_TYPES
from nnsmith.util import mkdir

from experiments.graphfuzz import GraphFuzz

import pickle
import os
import random
import argparse
import time
import warnings
import tarfile

from tqdm import tqdm
import torch


def nnsmith_gen_once(
    path_prefix, seed, max_nodes, candidates_overwrite=None, mode="random"
):
    if mode == "hybrid":
        mode = random.choice(["random", "guided"])

    torch.manual_seed(seed)
    gen_tstart = time.time()
    gen, solution = random_model_gen(
        init_rank=4,
        seed=seed,
        max_nodes=max_nodes,
        candidates_overwrite=candidates_overwrite,
        mode=mode,
    )
    net = SymbolNet(
        gen.abstract_graph, solution, verbose=False, alive_shapes=gen.alive_shapes
    )
    gen_time = time.time() - gen_tstart

    net.enable_proxy_grad()
    net.eval()  # otherwise BN wants batch > 1
    searcher = PracticalHybridSearch(net)
    n_try, sat_inputs = searcher.search(
        max_time_ms=gen_time * 0.02 * 1000, max_sample=2, return_list=True
    )
    net.disable_proxy_grad()

    with torch.no_grad():
        net.eval()

        test_inputs = sat_inputs if sat_inputs else net.get_random_inps(use_cuda=False)

        outputs = net.forward(*test_inputs)

        inames, onames, oidx = torch2onnx(
            net,
            path_prefix + ".onnx",
            verbose=False,
            use_cuda=False,
            dummy_inputs=test_inputs,
        )

        inputs = [t.cpu().numpy() for t in test_inputs]

        if isinstance(outputs, torch.Tensor):
            outputs = [outputs.cpu().numpy()]
        else:
            outputs = [o.cpu().numpy() for o in outputs]

        input_dict = {ina: inp for ina, inp in zip(inames, inputs)}
        output_dict = {oname: outputs[i] for oname, i in zip(onames, oidx)}

        with open(path_prefix + ".pkl", "wb") as f:
            pickle.dump((input_dict, output_dict), f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--onnx_dir", type=str, required=True)
    parser.add_argument("--time_budget", type=int, default=60 * 60 * 4)
    parser.add_argument("--max_nodes", type=int, default=10)
    parser.add_argument("--graphfuzz_ops", action="store_true")
    parser.add_argument("--ort_cache", type=str, default=None)
    parser.add_argument("--seed", type=int, default=233)
    parser.add_argument("--mode", type=str, default="random")
    parser.add_argument("--tar", action="store_true")
    args = parser.parse_args()

    mkdir(args.onnx_dir)

    random.seed(args.seed)
    torch.manual_seed(args.seed)

    if args.ort_cache:
        print(args.ort_cache)
        if not os.path.exists(args.ort_cache):
            print(f"Please first generate cache! (mkdir config first)")
            print(f"python nnsmith/dtype_test.py --cache {args.ort_cache}")
            exit(1)
        # must pre run this. otherwise using ort will slow down generation.
        rewrite_op_dtype(ALL_OP_TYPES, factory=None, cache=args.ort_cache)

    if args.graphfuzz_ops:
        candidates_overwrite = GraphFuzz.get_available_op_ts()
    else:
        candidates_overwrite = None

    # FORMAT: {generation time cost in seconds}, {model relative path}
    # MUST RANK by GENERATION ORDER.
    config_file = open(os.path.join(args.onnx_dir, "gentime.csv"), "w")

    start_time = time.time()
    gen_cnt = 0
    valid_cnt = 0

    if args.tar:
        tar = tarfile.open(os.path.join(args.onnx_dir, "models.tar"), "w")

    with tqdm(total=args.time_budget) as pbar:
        while time.time() - start_time < args.time_budget:
            seed = random.getrandbits(32)

            tstart = time.time()
            try:
                with warnings.catch_warnings():  # just shutup.
                    warnings.simplefilter("ignore")
                    nnsmith_gen_once(
                        os.path.join(args.onnx_dir, f"{valid_cnt}"),
                        seed,
                        max_nodes=10,
                        candidates_overwrite=candidates_overwrite,
                        mode=args.mode,
                    )
                to_name = f"{valid_cnt}.onnx"
                label = to_name
                valid_cnt += 1
                if args.tar:
                    tar.add(
                        os.path.join(args.onnx_dir, to_name),
                        arcname="models/" + to_name,
                    )
                    os.unlink(os.path.join(args.onnx_dir, to_name))
            except Exception as e:
                print(f"Fail when seed={seed}")
                print(e)  # Skip a few errors.
                label = "FAILURE"

            time_diff = time.time() - tstart
            config_file.write(f"{time_diff:.5f},{label}\n")

            gen_cnt += 1
            config_file.flush()

            pbar.update(int(time.time() - start_time) - pbar.n)
            pbar.set_description(f"valid={valid_cnt},fail={gen_cnt-valid_cnt}")
            pbar.refresh()
        config_file.close()



import argparse
import matplotlib.pyplot as plt
import pandas
import os


class Ploter:
    def __init__(self, cov_lim=None) -> None:
        self.legends = []  # type: ignore
        # cov / time, cov / iteration, iteration / time
        fig, axs = plt.subplots(1, 3, constrained_layout=True, figsize=(16, 5))
        self.fig = fig
        self.axs = axs
        self.cov_lim = cov_lim

    def add(self, folder, name=None):
        path = os.path.join(folder, "cov_by_time.csv")
        df = pandas.read_csv(path, usecols=[0, 1], header=None).to_numpy()

        self.axs[0].plot(df[:, 0], df[:, 1])  # cov / time
        self.axs[1].plot(range(len(df[:, 0])), df[:, 1])  # cov / iteration
        self.axs[2].plot(df[:, 0], range(len(df[:, 1])))  # iter / time

        if name:
            self.legends.append(name)
        else:
            assert not self.legends

    def plot(self, save="cov"):
        for axs in self.axs:
            axs.legend(self.legends)
        # plt.legend(self.legends)

        if self.cov_lim is not None:
            self.axs[0].set_ylim(bottom=self.cov_lim)
            self.axs[1].set_ylim(bottom=self.cov_lim)

        self.axs[0].set(xlabel="Time / Second", ylabel="# Coverage")
        self.axs[0].set_title("Coverage $\\bf{Time}$ Efficiency")

        self.axs[1].set(ylabel="# Coverage", xlabel="# Iteration")
        self.axs[1].set_title("Coverage $\\bf{Iteration}$ Efficiency")

        self.axs[2].set(xlabel="Time / Second", ylabel="# Iteration")
        self.axs[2].set_title("Iteration Speed")

        plt.savefig(save + ".pdf")
        plt.savefig(save + ".png")


if "__main__" == __name__:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f", "--folders", type=str, nargs="+", help="bug report folder"
    )
    parser.add_argument(
        "-cl", "--cov_lim", type=int, default=None, help="coverage starting lim"
    )
    parser.add_argument(
        "--tvmfuzz", type=str, nargs="?", help="TVMFuzz coverage by time file"
    )
    args = parser.parse_args()

    ploter = Ploter(cov_lim=args.cov_lim)

    for f in args.folders:
        ploter.add(f, f)
    ploter.plot("cov")
